"""
Interface with a Vault (or OpenBao) server and the KV secret backend.

.. important::
    This module requires the general :ref:`Vault setup <vault-setup>`.
"""

import logging

from salt.defaults import NOT_SET
from salt.exceptions import CommandExecutionError
from salt.exceptions import SaltException
from salt.exceptions import SaltInvocationError

from saltext.vault.utils import vault
from saltext.vault.utils.versions import warn_until

log = logging.getLogger(__name__)


def read_secret(path, key=None, metadata=False, default=NOT_SET, version=None):
    """
    Return the value of <key> at <path> in vault, or entire secret.

    CLI Example:

    .. code-block:: bash

        salt '*' vault.read_secret salt/kv/secret

    Required policy:

    .. code-block:: vaultpolicy

        path "<mount>/<secret>" {
            capabilities = ["read"]
        }

        # or KV v2
        path "<mount>/data/<secret>" {
            capabilities = ["read"]
        }

    path
        The path to the secret, including mount.

    key
        The data field at <path> to read. If unspecified, returns the
        whole dataset.

    metadata
        If using KV v2 backend, display full results, including metadata.
        Defaults to False.

    default
        When the path or path/key combination is not found, an exception will
        be raised, unless a default is provided here.

    version
        The version to read. If unset, reads the latest one.

        .. versionadded:: 1.2.0
    """
    if default == NOT_SET:
        default = CommandExecutionError
    if key is not None:
        metadata = False
    log.debug("Reading Vault secret for %s at %s", __grains__.get("id"), path)
    try:
        data = vault.read_kv(
            path, __opts__, __context__, include_metadata=metadata, version=version
        )
        if key is not None:
            return data[key]
        return data
    except Exception as err:  # pylint: disable=broad-except
        if default is CommandExecutionError:
            raise CommandExecutionError(
                f"Failed to read secret! {type(err).__name__}: {err}"
            ) from err
        return default


def read_secret_meta(path):
    """
    .. versionadded:: 1.2.0

    Return secret metadata and versions for <path>.
    Requires KV v2.

    CLI Example:

    .. code-block:: bash

        salt '*' vault.read_secret_meta salt/kv/secret

    Required policy:

    .. code-block:: vaultpolicy

        path "<mount>/metadata/<secret>" {
            capabilities = ["read"]
        }

    path
        The path to the secret, including mount.
    """
    log.debug("Reading Vault secret metadata for %s at %s", __grains__.get("id"), path)
    try:
        return vault.read_kv_meta(path, __opts__, __context__)
    except Exception as err:  # pylint: disable=broad-except
        log.error("Failed to read secret metadata! %s: %s", type(err).__name__, err)
        return False


def write_secret(path, **kwargs):
    """
    Set secret dataset at <path>.
    Fields are specified as arbitrary keyword arguments.

    CLI Example:

    .. code-block:: bash

            salt '*' vault.write_secret "secret/my/secret" user="foo" password="bar"

    Required policy:

    .. code-block:: vaultpolicy

        path "<mount>/<secret>" {
            capabilities = ["create", "update"]
        }

        # or KV v2
        path "<mount>/data/<secret>" {
            capabilities = ["create", "update"]
        }

    path
        The path to the secret, including mount.
    """
    log.debug("Writing vault secrets for %s at %s", __grains__.get("id"), path)
    data = {x: y for x, y in kwargs.items() if not x.startswith("__")}
    try:
        res = vault.write_kv(path, data, __opts__, __context__)
        if isinstance(res, dict):
            return res["data"]
        return res
    except Exception as err:  # pylint: disable=broad-except
        log.error("Failed to write secret! %s: %s", type(err).__name__, err)
        return False


def write_raw(path, raw):
    """
    Set raw data at <path>.

    CLI Example:

    .. code-block:: bash

            salt '*' vault.write_raw "secret/my/secret" '{"user":"foo","password": "bar"}'

    Required policy: see :func:`write_secret`

    path
        The path to the secret, including mount.

    raw
        Secret data to write to <path>. Has to be a mapping.
    """
    log.debug("Writing vault secrets for %s at %s", __grains__.get("id"), path)
    try:
        res = vault.write_kv(path, raw, __opts__, __context__)
        if isinstance(res, dict):
            return res["data"]
        return res
    except Exception as err:  # pylint: disable=broad-except
        log.error("Failed to write secret! %s: %s", type(err).__name__, err)
        return False


def patch_secret(path, **kwargs):
    """
    Patch secret dataset at <path>. Fields are specified as arbitrary keyword arguments.

    .. note::

        This works even for older Vault versions, KV v1 and with missing
        ``patch`` capability, but will use more than one request to simulate
        the functionality by issuing a read and update request.

        For proper, single-request patching, requires versions of KV v2 that
        support the ``patch`` capability and the ``patch`` capability to be available
        for the path.

    .. note::

        This uses JSON Merge Patch format internally.
        Keys set to ``null`` (JSON/YAML)/``None`` (Python) will be deleted.

    CLI Example:

    .. code-block:: bash

            salt '*' vault.patch_secret "secret/my/secret" password="baz"

    Required policy:

    .. code-block:: vaultpolicy

        # Proper patching
        path "<mount>/data/<secret>" {
            capabilities = ["patch"]
        }

        # OR (!), for older KV v2 setups:

        path "<mount>/data/<secret>" {
            capabilities = ["read", "update"]
        }

        # OR (!), for KV v1 setups:

        path "<mount>/<secret>" {
            capabilities = ["read", "update"]
        }

    path
        The path to the secret, including mount.
    """
    log.debug("Patching vault secrets for %s at %s", __grains__.get("id"), path)
    data = {x: y for x, y in kwargs.items() if not x.startswith("__")}
    try:
        res = vault.patch_kv(path, data, __opts__, __context__)
        if isinstance(res, dict):
            return res["data"]
        return res
    except Exception as err:  # pylint: disable=broad-except
        log.error("Failed to patch secret! %s: %s", type(err).__name__, err)
        return False


def delete_secret(path, *args, **kwargs):
    """
    Delete secret at <path>. If <path> is on KV v2, the secret will be soft-deleted.

    CLI Example:

    .. code-block:: bash

        salt '*' vault.delete_secret "secret/my/secret"
        salt '*' vault.delete_secret "secret/my/secret" 1 2 3
        salt '*' vault.delete_secret "secret/my/secret" all_versions=true

    Required policy:

    .. code-block:: vaultpolicy

        path "<mount>/<secret>" {
            capabilities = ["delete"]
        }

        # or KV v2
        path "<mount>/data/<secret>" {
            capabilities = ["delete"]
        }

        # KV v2 versions
        # all_versions=True additionally requires the policy for vault.read_secret_meta
        path "<mount>/delete/<secret>" {
            capabilities = ["update"]
        }

    path
        The path to the secret, including mount.

    all_versions
        .. versionadded:: 1.2.0

        Delete all versions of the secret for KV v2.
        Can only be passed as a keyword argument.
        Defaults to false.

    .. versionadded:: 1.0.0

        For KV v2, you can specify versions to soft-delete as supplemental
        positional arguments.
    """
    all_versions = kwargs.pop("all_versions", False)
    unknown_kwargs = tuple(x for x in kwargs if not x.startswith("_"))
    if unknown_kwargs:
        raise SaltInvocationError(f"Passed unknown keyword arguments: {' '.join(unknown_kwargs)}")
    log.debug("Deleting vault secrets for %s in %s", __grains__.get("id"), path)
    if args:
        log.debug(f"Affected versions: {' '.join(str(x) for x in args)}")
    try:
        return vault.delete_kv(
            path, __opts__, __context__, versions=list(args) or None, all_versions=all_versions
        )
    except Exception as err:  # pylint: disable=broad-except
        log.error("Failed to delete secret! %s: %s", type(err).__name__, err)
        return False


def restore_secret(path, *versions, **kwargs):
    """
    .. versionadded:: 1.2.0

    Restore specific versions of a secret path. Only supported on Vault KV v2.

    CLI Example:

    .. code-block:: bash

        salt '*' vault.restore_secret secret/my/secret 1 2

    Required policy:

    .. code-block:: vaultpolicy

        # all_versions=True or defaulting to the most recent version additionally
        # requires the policy for vault.read_secret_meta
        path "<mount>/undelete/<secret>" {
            capabilities = ["update"]
        }

    path
        The path to the secret, including mount.

    all_versions
        Restore all versions of the secret for KV v2.
        Can only be passed as a keyword argument.
        Defaults to false.

    You can specify versions to restore as supplemental positional arguments.
    If no version is specified, tries to restore the latest version, and if
    the latest version has not been deleted, fails.
    """
    all_versions = kwargs.pop("all_versions", False)
    unknown_kwargs = tuple(x for x in kwargs if not x.startswith("_"))
    if unknown_kwargs:
        raise SaltInvocationError(f"Passed unknown keyword arguments: {' '.join(unknown_kwargs)}")
    log.debug("Restoring vault secrets for %s in %s", __grains__.get("id"), path)

    try:
        return vault.restore_kv(
            path, __opts__, __context__, list(versions) or None, all_versions=all_versions
        )
    except vault.VaultException as err:
        raise CommandExecutionError(f"{err.__class__.__name__}: {err}") from err


def destroy_secret(path, *args, **kwargs):
    """
    Destroy specified secret versions at <path>. Only supported on Vault KV v2.

    CLI Example:

    .. code-block:: bash

        salt '*' vault.destroy_secret "secret/my/secret"
        salt '*' vault.destroy_secret "secret/my/secret" 1 2
        salt '*' vault.destroy_secret "secret/my/secret" all_versions=true

    Required policy:

    .. code-block:: vaultpolicy

        # all_versions=True or defaulting to the most recent version additionally
        # requires the policy for vault.read_secret_meta
        path "<mount>/destroy/<secret>" {
            capabilities = ["update"]
        }

    path
        The path to the secret, including mount.

    all_versions
        .. versionadded:: 1.2.0

        Delete all versions of the secret for KV v2.
        Can only be passed as a keyword argument.
        Defaults to false.

    You can specify versions to destroy as supplemental positional arguments.

    .. versionchanged:: 1.2.0

        If no version was specified, defaults to the most recent one.
    """
    all_versions = kwargs.pop("all_versions", False)
    unknown_kwargs = tuple(x for x in kwargs if not x.startswith("_"))
    if unknown_kwargs:
        raise SaltInvocationError(f"Passed unknown keyword arguments: {' '.join(unknown_kwargs)}")
    log.debug("Destroying vault secrets for %s in %s", __grains__.get("id"), path)
    if args:
        log.debug(f"Affected versions: {' '.join(str(x) for x in args)}")
    try:
        return vault.destroy_kv(
            path, list(args) or None, __opts__, __context__, all_versions=all_versions
        )
    except Exception as err:  # pylint: disable=broad-except
        log.error("Failed to destroy secret! %s: %s", type(err).__name__, err)
        return False


def wipe_secret(path):
    """
    .. versionadded:: 1.2.0

    Remove all version history and data for the secret at <path>.
    Requires KV v2.

    CLI Example:

    .. code-block:: bash

        salt '*' vault.wipe_secret "secret/my/secret"

    Required policy:

    .. code-block:: vaultpolicy

        path "<mount>/metadata/<secret>" {
            capabilities = ["delete"]
        }
    """
    log.debug("Wiping vault secrets for %s in %s", __grains__.get("id"), path)
    try:
        return vault.wipe_kv(path, __opts__, __context__)
    except Exception as err:  # pylint: disable=broad-except
        log.error("Failed to wipe secret! %s: %s", type(err).__name__, err)
        return False


def list_secrets(path, default=NOT_SET, keys_only=None):
    """
    List secret keys at <path>. The path should end with a trailing slash.

    CLI Example:

    .. code-block:: bash

        salt '*' vault.list_secrets "secret/my/"

    Required policy:

    .. code-block:: vaultpolicy

        path "<mount>/<path>" {
            capabilities = ["list"]
        }

        # or KV v2
        path "<mount>/metadata/<path>" {
            capabilities = ["list"]
        }

    path
        The path to the secret, including mount.

    default
        When the path is not found, an exception will be raised, unless a default
        is provided here.

    keys_only
        .. versionadded:: 1.0.0

        This function used to return a dictionary like ``{"keys": ["some/", "some/key"]}``.
        Setting this to True will only return the list of keys.
        For backwards-compatibility reasons, this currently defaults to False.
        Beginning with version 2 of this extension, the default will change to True.
    """
    if default == NOT_SET:
        default = CommandExecutionError
    if keys_only is None:
        try:
            warn_until(
                2,
                (
                    "In version {version}, this function will return the list of "
                    "secret keys only. You can switch to the new behavior explicitly "
                    "by specifying keys_only=True."
                ),
            )
            keys_only = False
        except RuntimeError:
            keys_only = True

    log.debug("Listing vault secret keys for %s in %s", __grains__.get("id"), path)
    try:
        keys = vault.list_kv(path, __opts__, __context__)
        if keys_only:
            return keys
        # this is the way Salt behaved previously
        return {"keys": keys}
    except Exception as err:  # pylint: disable=broad-except
        if default is CommandExecutionError:
            raise CommandExecutionError(
                f"Failed to list secrets! {type(err).__name__}: {err}"
            ) from err
        return default


def clear_cache(connection=True, session=False):
    """
    .. versionadded:: 1.0.0

    Delete Vault caches. Will ensure the current token and associated leases
    are revoked by default.

    The cache is organized in a hierarchy: ``/vault/connection/session/leases``.
    (*italics* mark data that is only cached when receiving configuration from a master)

    ``connection`` contains KV metadata (by default), *configuration* and *(AppRole) auth credentials*.
    ``session`` contains the currently active token.
    ``leases`` contains leases issued to the currently active token like database credentials.

    CLI Example:

    .. code-block:: bash

        salt '*' vault.clear_cache
        salt '*' vault.clear_cache session=True

    connection
        Only clear the cached data scoped to a connection. This includes
        configuration, auth credentials, the currently active auth token
        as well as leases and KV metadata (by default). Defaults to true.
        Set this to false to clear all Vault caches.

    session
        Only clear the cached data scoped to a session. This only includes
        leases and the currently active auth token, but not configuration
        or (AppRole) auth credentials. Defaults to false.
        Setting this to true will keep the connection cache, regardless
        of ``connection``.
    """
    return vault.clear_cache(__opts__, __context__, connection=connection, session=session)


def clear_token_cache():
    """
    .. deprecated:: 1.0.0
    .. versionchanged:: 1.0.0

        This is now an alias for :func:`vault.clear_cache<clear_cache>` with ``connection=True``
        and ``session=False`` (the defaults).

    Delete minion Vault token cache.

    CLI Example:

    .. code-block:: bash

        salt '*' vault.clear_token_cache
    """
    log.debug("Deleting vault connection cache.")
    return clear_cache(connection=True, session=False)


def policy_fetch(policy):
    """
    .. versionadded:: 1.0.0

    Fetch the rules associated with an ACL policy. Returns ``None`` if the policy
    does not exist.

    CLI Example:

    .. code-block:: bash

        salt '*' vault.policy_fetch salt_minion

    Required policy:

    .. code-block:: vaultpolicy

        path "sys/policy/<policy>" {
            capabilities = ["read"]
        }

    policy
        The name of the policy to fetch.
    """
    # there is also "sys/policies/acl/{policy}"
    endpoint = f"sys/policy/{policy}"

    try:
        data = vault.query("GET", endpoint, __opts__, __context__)
        return data["rules"]

    except vault.VaultNotFoundError:
        return None
    except SaltException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err


def policy_write(policy, rules):
    r"""
    .. versionadded:: 1.0.0

    Create or update an ACL policy.

    CLI Example:

    .. code-block:: bash

        salt '*' vault.policy_write salt_minion 'path "secret/foo" {...}'

    Required policy:

    .. code-block:: vaultpolicy

        path "sys/policy/<policy>" {
            capabilities = ["create", "update"]
        }

    policy
        The name of the policy to create/update.

    rules
        Rules to write, formatted as in-line HCL.
    """
    endpoint = f"sys/policy/{policy}"
    payload = {"policy": rules}
    try:
        return vault.query(
            "POST", endpoint, __opts__, __context__, payload=payload, safe_to_retry=True
        )
    except SaltException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err


def policy_delete(policy):
    """
    .. versionadded:: 1.0.0

    Delete an ACL policy. Returns False if the policy does not exist.

    CLI Example:

    .. code-block:: bash

        salt '*' vault.policy_delete salt_minion

    Required policy:

    .. code-block:: vaultpolicy

        path "sys/policy/<policy>" {
            capabilities = ["delete"]
        }

    policy
        The name of the policy to delete.
    """
    endpoint = f"sys/policy/{policy}"

    try:
        return vault.query("DELETE", endpoint, __opts__, __context__)
    except vault.VaultNotFoundError:
        return False
    except SaltException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err


def policies_list():
    """
    .. versionadded:: 1.0.0

    List all ACL policies.

    CLI Example:

    .. code-block:: bash

        salt '*' vault.policies_list

    Required policy:

    .. code-block:: vaultpolicy

        path "sys/policy" {
            capabilities = ["read"]
        }
    """
    try:
        return vault.query("GET", "sys/policy", __opts__, __context__)["policies"]
    except SaltException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err


def query(method, endpoint, payload=None):
    """
    .. versionadded:: 1.0.0

    Issue arbitrary queries against the Vault API.

    CLI Example:

    .. code-block:: bash

        salt '*' vault.query GET auth/token/lookup-self

    Required policy: Depends on the query.

    You can ask the Vault CLI to output the necessary policy:

    .. code-block:: bash

        vault read -output-policy auth/token/lookup-self

    method
        HTTP method to use.

    endpoint
        Vault API endpoint to issue the request against. Do not include ``/v1/``.

    payload
        Optional dictionary to use as JSON payload.
    """
    try:
        return vault.query(method, endpoint, __opts__, __context__, payload=payload)
    except SaltException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err


def update_config(keep_session=False):
    """
    .. versionadded:: 1.0.0

    Attempt to update the cached configuration without clearing the
    currently active Vault session.

    CLI Example:

    .. code-block:: bash

        salt '*' vault.update_config

    keep_session
        Only update configuration that can be updated without
        creating a new login session.
        If this is false, still tries to keep the active session,
        but might clear it if the server configuration has changed
        significantly.
        Defaults to False.
    """
    return vault.update_config(__opts__, __context__, keep_session=keep_session)


def get_server_config():
    """
    .. versionadded:: 1.0.0

    Return the server connection configuration that's currently in use by Salt.
    Contains :vconf:`url <server:url>`, :vconf:`verify <server:verify>`
    and :vconf:`namespace <server:namespace>`.

    CLI Example:

    .. code-block:: bash

        salt '*' vault.get_server_config
    """
    try:
        client = vault.get_authd_client(__opts__, __context__)
        return client.get_config()
    except SaltException as err:
        raise CommandExecutionError(f"{type(err).__name__}: {err}") from err



import time
from datetime import datetime
from pathlib import Path

import pytest
from saltfactories.utils import random_string

from tests.support.mysql import MySQLImage
from tests.support.mysql import create_mysql_combo  # pylint: disable=unused-import
from tests.support.mysql import mysql_combo  # pylint: disable=unused-import
from tests.support.mysql import mysql_container  # pylint: disable=unused-import
from tests.support.vault import vault_delete
from tests.support.vault import vault_disable_secret_engine
from tests.support.vault import vault_enable_secret_engine
from tests.support.vault import vault_list
from tests.support.vault import vault_revoke
from tests.support.vault import vault_write

pytest.importorskip("docker")

pytestmark = [
    pytest.mark.skip_if_binaries_missing("vault"),
    pytest.mark.usefixtures("container"),
]


@pytest.fixture(scope="module")
def master_config_overrides():
    """
    You can override the default configuration per package by overriding this
    fixture in a conftest.py file.
    """
    return {
        "vault": {
            "cache": {
                "backend": "disk",
            },
            "policies": {
                "assign": [
                    "salt_minion",
                    "database_admin",
                ],
            },
        }
    }


@pytest.fixture(scope="module")
def mysql_image():
    version = "10.3"
    return MySQLImage(
        name="mariadb",
        tag=version,
        container_id=random_string(f"mariadb-{version}-"),
    )


@pytest.fixture
def role_args_common():
    return {
        "db_name": "testdb",
        "creation_statements": r"CREATE USER '{{name}}'@'%' IDENTIFIED BY '{{password}}';GRANT SELECT ON *.* TO '{{name}}'@'%';",
    }


@pytest.fixture
def testrole():
    return {
        "default_ttl": 3600,
        "max_ttl": 86400,
    }


@pytest.fixture
def testreissuerole():
    return {
        "default_ttl": 180,
        "max_ttl": 180,
    }


@pytest.fixture
def teststaticrole(mysql_container):
    return {
        "db_name": "testdb",
        "rotation_period": 86400,
        "username": mysql_container.mysql_user,
    }


@pytest.fixture
def testdb(mysql_container, container_host_ref):
    return {
        "plugin_name": "mysql-database-plugin",
        "connection_url": f"{{{{username}}}}:{{{{password}}}}@tcp({container_host_ref}:{mysql_container.mysql_port})/",
        "allowed_roles": "testrole,teststaticrole,testreissuerole",
        "username": "root",
        "password": mysql_container.mysql_passwd,
    }


@pytest.fixture(scope="module", autouse=True)
def db_engine(container):  # pylint: disable=unused-argument
    assert vault_enable_secret_engine("database")
    yield
    assert vault_disable_secret_engine("database")


@pytest.fixture
def connection_setup(testdb):
    try:
        vault_write("database/config/testdb", **testdb)
        assert "testdb" in vault_list("database/config")
        yield
    finally:
        # prevent dangling leases, which prevent disabling the secret engine
        assert vault_revoke("database/creds", prefix=True)
        if "testdb" in vault_list("database/config"):
            vault_delete("database/config/testdb")
            assert "testdb" not in vault_list("database/config")


@pytest.fixture(params=[["testrole"]])
def roles_setup(connection_setup, request, role_args_common):  # pylint: disable=unused-argument
    try:
        for role_name in request.param:
            role_args = request.getfixturevalue(role_name)
            role_args.update(role_args_common)
            vault_write(f"database/roles/{role_name}", **role_args)
            assert role_name in vault_list("database/roles")
        yield
    finally:
        for role_name in request.param:
            if role_name in vault_list("database/roles"):
                vault_delete(f"database/roles/{role_name}")
                assert role_name not in vault_list("database/roles")


@pytest.fixture
def role_static_setup(connection_setup, teststaticrole):  # pylint: disable=unused-argument
    role_name = "teststaticrole"
    try:
        vault_write(f"database/static-roles/{role_name}", **teststaticrole)
        assert role_name in vault_list("database/static-roles")
        yield
    finally:
        if role_name in vault_list("database/static-roles"):
            vault_delete(f"database/static-roles/{role_name}")
            assert role_name not in vault_list("database/static-roles")


@pytest.fixture(autouse=True)
def _cleanup():
    try:
        yield
    finally:
        # prevent dangling leases, which prevent disabling the secret engine
        assert vault_revoke("database/creds", prefix=True)
        if "testdb" in vault_list("database/config"):
            vault_delete("database/config/testdb")
            assert "testdb" not in vault_list("database/config")
        if "testrole" in vault_list("database/roles"):
            vault_delete("database/roles/testrole")
            assert "testrole" not in vault_list("database/roles")
        if "teststaticrole" in vault_list("database/static-roles"):
            vault_delete("database/static-roles/teststaticrole")
            assert "teststaticrole" not in vault_list("database/static-roles")


@pytest.mark.usefixtures("connection_setup")
def test_list_connections(salt_ssh_cli):
    ret = salt_ssh_cli.run("vault_db.list_connections")
    assert ret.returncode == 0
    assert ret.data == ["testdb"]


@pytest.mark.usefixtures("connection_setup")
def test_fetch_connection(salt_ssh_cli, testdb):
    ret = salt_ssh_cli.run("vault_db.fetch_connection", "testdb")
    assert ret.returncode == 0
    for var, val in testdb.items():
        if var == "password":
            continue
        if var in ["connection_url", "username"]:
            assert var in ret.data["connection_details"]
            assert ret.data["connection_details"][var] == val
        else:
            assert var in ret.data
            if var == "allowed_roles":
                assert ret.data[var] == list(val.split(","))
            else:
                assert ret.data[var] == val


@pytest.mark.usefixtures("testdb")
def test_write_connection(salt_ssh_cli, mysql_container, container_host_ref):
    args = {
        "plugin": "mysql",
        "connection_url": f"{{{{username}}}}:{{{{password}}}}@tcp({container_host_ref}:{mysql_container.mysql_port})/",
        "allowed_roles": ["testrole", "teststaticrole"],
        "username": "root",
        "password": mysql_container.mysql_passwd,
        # Can't rotate because we wouldn't know the new one for further tests
        "rotate": False,
    }
    ret = salt_ssh_cli.run("vault_db.write_connection", "testdb", **args)
    assert ret.returncode == 0
    assert "testdb" in vault_list("database/config")


@pytest.mark.usefixtures("connection_setup")
def test_update_connection(salt_ssh_cli):
    """
    Ensure missing kwargs are not enforced on updates.
    """
    ret = salt_ssh_cli.run(
        "vault_db.write_connection", "testdb", "mysql", allowed_roles=["*"], rotate=False
    )
    assert ret.returncode == 0
    assert ret.data is True


@pytest.mark.usefixtures("connection_setup")
def test_delete_connection(salt_ssh_cli):
    ret = salt_ssh_cli.run("vault_db.delete_connection", "testdb")
    assert ret.returncode == 0
    assert ret.data
    assert "testdb" not in vault_list("database/config")


@pytest.mark.usefixtures("connection_setup")
def test_reset_connection(salt_ssh_cli):
    ret = salt_ssh_cli.run("vault_db.reset_connection", "testdb")
    assert ret.returncode == 0
    assert ret.data


@pytest.mark.usefixtures("roles_setup")
def test_list_roles(salt_ssh_cli):
    ret = salt_ssh_cli.run("vault_db.list_roles")
    assert ret.returncode == 0
    assert ret.data == ["testrole"]


@pytest.mark.usefixtures("role_static_setup")
def test_list_roles_static(salt_ssh_cli):
    ret = salt_ssh_cli.run("vault_db.list_roles", static=True)
    assert ret.returncode == 0
    assert ret.data == ["teststaticrole"]


@pytest.mark.usefixtures("roles_setup")
def test_fetch_role(salt_ssh_cli, testrole):
    ret = salt_ssh_cli.run("vault_db.fetch_role", "testrole")
    assert ret.returncode == 0
    assert ret.data
    for var, val in testrole.items():
        assert var in ret.data
        if var == "creation_statements":
            assert ret.data[var] == [val]
        else:
            assert ret.data[var] == val


@pytest.mark.usefixtures("role_static_setup")
def test_fetch_role_static(salt_ssh_cli, teststaticrole):
    ret = salt_ssh_cli.run("vault_db.fetch_role", "teststaticrole", static=True)
    assert ret.returncode == 0
    assert ret.data
    for var, val in teststaticrole.items():
        assert var in ret.data
        assert ret.data[var] == val


@pytest.mark.usefixtures("connection_setup")
def test_write_role(salt_ssh_cli):
    args = {
        "connection": "testdb",
        "creation_statements": r"CREATE USER '{{name}}'@'%' IDENTIFIED BY '{{password}}';GRANT SELECT ON *.* TO '{{name}}'@'%';",
    }
    ret = salt_ssh_cli.run("vault_db.write_role", "testrole", **args)
    assert ret.returncode == 0
    assert ret.data
    assert "testrole" in vault_list("database/roles")


@pytest.mark.usefixtures("connection_setup")
def test_write_static_role(salt_ssh_cli, mysql_container):
    args = {
        "connection": "testdb",
        "username": mysql_container.mysql_user,
        "rotation_period": 86400,
    }
    ret = salt_ssh_cli.run("vault_db.write_static_role", "teststaticrole", **args)
    assert ret.returncode == 0
    assert ret.data
    assert "teststaticrole" in vault_list("database/static-roles")


@pytest.mark.usefixtures("roles_setup")
def test_delete_role(salt_ssh_cli):
    ret = salt_ssh_cli.run("vault_db.delete_role", "testrole")
    assert ret.returncode == 0
    assert ret.data
    assert "testrole" not in vault_list("database/roles")


@pytest.mark.usefixtures("role_static_setup")
def test_delete_role_static(salt_ssh_cli):
    ret = salt_ssh_cli.run("vault_db.delete_role", "teststaticrole", static=True)
    assert ret.returncode == 0
    assert ret.data
    assert "teststaticrole" not in vault_list("database/static-roles")


@pytest.fixture(params=({},))
def _cached_creds(salt_ssh_cli, roles_setup, request):  # pylint: disable=unused-argument
    data = request.param.copy()
    role = data.pop("role", "testrole")
    ret = salt_ssh_cli.run("vault_db.get_creds", role, cache=True, **data)
    assert ret.returncode == 0
    assert "username" in ret.data
    assert "password" in ret.data
    yield ret.data
    # We need to get rid of the cached data since the lease is
    # revoked after each test and we don't run with check_server.
    ret = salt_ssh_cli.run("vault_db.clear_cached")
    assert ret.returncode == 0
    assert ret.data is True


@pytest.mark.usefixtures("roles_setup")
def test_get_creds(salt_ssh_cli):
    ret = salt_ssh_cli.run("vault_db.get_creds", "testrole", cache=False)
    assert ret.returncode == 0
    assert "username" in ret.data
    assert "password" in ret.data


@pytest.mark.usefixtures("role_static_setup")
def test_get_creds_static(salt_ssh_cli, teststaticrole):
    ret = salt_ssh_cli.run("vault_db.get_creds", "teststaticrole", static=True, cache=False)
    assert ret.returncode == 0
    assert "username" in ret.data
    assert "password" in ret.data
    assert ret.data["username"] == teststaticrole["username"]


def test_get_creds_cached(salt_ssh_cli, _cached_creds, master):
    ret_new = salt_ssh_cli.run("vault_db.get_creds", "testrole", cache=True)
    assert ret_new.returncode == 0
    assert "username" in ret_new.data
    assert "password" in ret_new.data
    assert ret_new.data["username"] == _cached_creds["username"]
    assert ret_new.data["password"] == _cached_creds["password"]
    # Ensure we're caching to the master per-minion cache,
    # not the Salt-SSH minion one. This allows master-side
    # renewals for Salt-SSH minions via an engine module.
    cachedir = Path(master.config["cachedir"])
    cache_file = (
        cachedir
        / "minions"
        / "localhost"
        / "vault"
        / "connection"
        / "session"
        / "leases"
        / "db.database.dynamic.testrole.default.p"
    )
    assert cache_file.exists()


@pytest.mark.usefixtures("_cached_creds")
def test_clear_cached(salt_ssh_cli):
    list_cached = salt_ssh_cli.run("vault_db.list_cached")
    assert list_cached.returncode == 0
    assert list_cached.data
    clear_cached = salt_ssh_cli.run("vault_db.clear_cached")
    assert clear_cached.returncode == 0
    assert clear_cached.data is True
    list_cached_new = salt_ssh_cli.run("vault_db.list_cached")
    assert list_cached_new.returncode == 0
    assert not list_cached_new.data


@pytest.mark.usefixtures("_cached_creds")
def test_list_cached(salt_ssh_cli):
    ret = salt_ssh_cli.run("vault_db.list_cached")
    assert ret.returncode == 0
    ckey = "db.database.dynamic.testrole.default"
    assert ret.data
    assert ckey in ret.data
    assert not ret.data[ckey]["expired"]
    assert ret.data[ckey]["expires_in"] > 3590
    assert "data" not in ret.data[ckey]
    now = datetime.now().astimezone()
    # this might fail if this test runs juuust before midnight
    assert ret.data[ckey]["creation_time"].startswith(now.strftime("%Y-%m-%d"))
    # I hope you have something better to do during New Year's Eve
    assert ret.data[ckey]["expire_time"].startswith(now.strftime("%Y-"))
    for val in ("creation_time", "expire_time"):
        assert ret.data[ckey][val].endswith(now.strftime(" %Z"))


@pytest.mark.usefixtures("_cached_creds")
def test_renew_cached(salt_ssh_cli):
    curr = salt_ssh_cli.run("vault_db.list_cached")
    assert curr.returncode == 0
    curr = curr.data
    ckey = "db.database.dynamic.testrole.default"
    assert curr
    assert ckey in curr
    time.sleep(3)
    renew_cached = salt_ssh_cli.run("vault_db.renew_cached")
    assert renew_cached.returncode == 0
    assert renew_cached.data is True
    new = salt_ssh_cli.run("vault_db.list_cached")
    assert new.returncode == 0
    assert new.data[ckey]["expire_time"] != curr[ckey]["expire_time"]


@pytest.mark.usefixtures("role_static_setup")
def test_rotate_static_role(salt_ssh_cli):
    ret = salt_ssh_cli.run("vault_db.get_creds", "teststaticrole", static=True, cache=False)
    assert ret.returncode == 0
    assert ret.data
    old_pw = ret.data["password"]
    ret = salt_ssh_cli.run("vault_db.rotate_static_role", "teststaticrole")
    assert ret.returncode == 0
    assert ret.data
    ret = salt_ssh_cli.run("vault_db.get_creds", "teststaticrole", static=True, cache=False)
    assert ret.returncode == 0
    assert ret.data["password"] != old_pw



import pytest

import saltext.vault.beacons.vault_lease as lease


@pytest.fixture
def configure_loader_modules():
    return {
        lease: {
            "__grains__": {"id": "test-minion"},
        }
    }


@pytest.mark.parametrize(
    "config,exp",
    (
        ({}, "Configuration for vault_lease must be a list"),
        ([], "Requires monitored lease(s) cache key(s) in `leases`"),
        ([{"leases": 123}], "`leases` must be a dict, list or str"),
        ([{"leases": "foo"}], True),
        ([{"leases": "foo.*"}], "`leases` does not support globs"),
        ([{"leases": ["foo", "bar"]}], True),
        ([{"leases": ["foo", "bar.*"]}], "`leases` does not support globs"),
        ([{"leases": {"foo": "foo", "bar": "bar"}}], "`leases` mapping values must be dicts"),
        ([{"leases": {"foo": {}, "bar": {}}}], True),
        ([{"leases": {"foo": {"min_ttl": "1d"}, "bar": {}}}], True),
        ([{"leases": {"foo": {"min_ttl": "1d"}, "bar.*": {}}}], "`leases` does not support globs"),
    ),
)
def test_validate(config, exp):
    res, msg = lease.validate(config)
    if exp is True:
        assert res is True
    else:
        assert msg == exp


@pytest.mark.parametrize(
    "config,exp",
    (
        ([{"leases": "foo"}], {"leases": {"foo": {}}}),
        (
            [{"leases": "foo"}, {"min_ttl": 42}, {"check_server": True}, {"meta": "foo"}],
            {"leases": {"foo": {"min_ttl": 42, "check_server": True, "meta": "foo"}}},
        ),
        ([{"leases": ["foo", "bar"]}], {"leases": {"foo": {}, "bar": {}}}),
        (
            [{"leases": ["foo", "bar"]}, {"min_ttl": 42}, {"check_server": True}, {"meta": "foo"}],
            {
                "leases": {
                    "foo": {"min_ttl": 42, "check_server": True, "meta": "foo"},
                    "bar": {"min_ttl": 42, "check_server": True, "meta": "foo"},
                }
            },
        ),
        ([{"leases": {"foo": {}, "bar": {}}}], {"leases": {"foo": {}, "bar": {}}}),
        (
            [
                {"leases": {"foo": {}, "bar": {}}},
                {"min_ttl": 42},
                {"check_server": True},
                {"meta": "foo"},
            ],
            {
                "leases": {
                    "foo": {"min_ttl": 42, "check_server": True, "meta": "foo"},
                    "bar": {"min_ttl": 42, "check_server": True, "meta": "foo"},
                }
            },
        ),
        (
            [
                {"leases": {"foo": {"min_ttl": 1337}, "bar": {"check_server": False}}},
                {"min_ttl": 42},
                {"check_server": True},
                {"meta": "foo"},
            ],
            {
                "leases": {
                    "foo": {"min_ttl": 1337, "check_server": True, "meta": "foo"},
                    "bar": {"min_ttl": 42, "check_server": False, "meta": "foo"},
                }
            },
        ),
        (
            [
                {"leases": {"foo": {"min_ttl": None}, "bar": {"meta": "bar"}}},
                {"min_ttl": 42},
                {"check_server": True},
                {"meta": "foo"},
            ],
            {
                "leases": {
                    "foo": {"min_ttl": None, "check_server": True, "meta": "foo"},
                    "bar": {"min_ttl": 42, "check_server": True, "meta": "bar"},
                }
            },
        ),
        (
            [
                {"leases": {"foo": {"meta": ["foo"]}, "bar": {"meta": {"bar": True}}}},
                {"meta": "foo"},
            ],
            {"leases": {"foo": {"meta": ["foo"]}, "bar": {"meta": {"bar": True}}}},
        ),
        (
            [
                {"leases": {"foo": {"meta": ["baz"]}, "bar": {"meta": {"bar": True}}}},
                {"meta": ["foo"]},
            ],
            {"leases": {"foo": {"meta": ["baz"]}, "bar": {"meta": {"bar": True}}}},
        ),
        (
            [
                {"leases": {"foo": {"meta": ["baz"]}, "bar": {"meta": {"bar": True}}}},
                {"meta": {"foo": True}},
            ],
            {"leases": {"foo": {"meta": ["baz"]}, "bar": {"meta": {"bar": True}}}},
        ),
        (
            [{"leases": ["foo", "bar"]}, {"renew": False}],
            {"leases": {"foo": {"renew": False}, "bar": {"renew": False}}},
        ),
        (
            [{"leases": {"foo": {"renew": True}, "bar": {}}}, {"renew": False}],
            {"leases": {"foo": {"renew": True}, "bar": {"renew": False}}},
        ),
    ),
)
def test_render_config(config, exp):
    res = lease._render_config(config)
    assert res == exp


@pytest.mark.parametrize(
    "cfg,info,exp",
    (
        ({}, {"min_ttl": 1234, "meta": None}, {"min_ttl": 1234, "meta": None}),
        ({}, {"min_ttl": "1h", "meta": None}, {"min_ttl": "1h", "meta": None}),
        ({"min_ttl": "1h"}, {"min_ttl": None, "meta": None}, {"min_ttl": "1h", "meta": None}),
        (
            {"check_server": True},
            {"min_ttl": None, "meta": None},
            {"min_ttl": 300, "meta": None, "check_server": True},
        ),
        (
            {"check_server": False},
            {"min_ttl": None, "meta": None},
            {"min_ttl": 300, "meta": None, "check_server": False},
        ),
        ({"min_ttl": "2h"}, {"min_ttl": 3600, "meta": None}, {"min_ttl": "2h", "meta": None}),
        ({"min_ttl": "1h"}, {"min_ttl": "2h", "meta": None}, {"min_ttl": "2h", "meta": None}),
        ({}, {"min_ttl": None, "meta": "foo"}, {"min_ttl": 300, "meta": "foo"}),
        ({"min_ttl": 42}, {"min_ttl": None, "meta": None}, {"min_ttl": 42, "meta": None}),
        ({"meta": 123}, {"min_ttl": None, "meta": None}, {"min_ttl": 300, "meta": 123}),
        ({"meta": ["foo"]}, {"min_ttl": None, "meta": None}, {"min_ttl": 300, "meta": ["foo"]}),
        (
            {"meta": {"foo": True}},
            {"min_ttl": None, "meta": None},
            {"min_ttl": 300, "meta": {"foo": True}},
        ),
        ({"meta": "foo"}, {"min_ttl": None, "meta": "bar"}, {"min_ttl": 300, "meta": "bar"}),
        ({"meta": ["foo"]}, {"min_ttl": None, "meta": "bar"}, {"min_ttl": 300, "meta": "bar"}),
        (
            {"meta": {"foo": True}},
            {"min_ttl": None, "meta": "bar"},
            {"min_ttl": 300, "meta": "bar"},
        ),
        (
            {"meta": {"foo": True}},
            {"min_ttl": None, "meta": ["bar"]},
            {"min_ttl": 300, "meta": ["bar"]},
        ),
        (
            {"meta": ["foo"]},
            {"min_ttl": None, "meta": {"bar": True}},
            {"min_ttl": 300, "meta": {"bar": True}},
        ),
        (
            {"meta": ["foo"]},
            {"min_ttl": None, "meta": ["bar"]},
            {"min_ttl": 300, "meta": ["foo", "bar"]},
        ),
        (
            {"meta": {"foo": True}},
            {"min_ttl": None, "meta": {"bar": True}},
            {"min_ttl": 300, "meta": {"foo": True, "bar": True}},
        ),
        (
            {"meta": {"foo": ["a"]}},
            {"min_ttl": None, "meta": {"foo": ["b"]}},
            {"min_ttl": 300, "meta": {"foo": ["a", "b"]}},
        ),
        (
            {"meta": {"foo": True}},
            {"min_ttl": None, "meta": {"foo": False}},
            {"min_ttl": 300, "meta": {"foo": False}},
        ),
    ),
)
def test_merge_lease_config(cfg, info, exp):
    res = lease._merge_lease_config(cfg, info)
    assert res == exp


@pytest.mark.parametrize(
    "cfg,info,exp",
    (
        (
            {"min_ttl": 1234},
            {"min_ttl": 42, "meta": None},
            {"ckey": "test.lease", "min_ttl": 1234, "meta": None, "check_server": None},
        ),
        (
            {"min_ttl": 300, "meta": "foo"},
            {"min_ttl": None, "meta": None},
            {"ckey": "test.lease", "min_ttl": 300, "meta": "foo", "check_server": None},
        ),
        (
            {"min_ttl": 300, "meta": ["foo"]},
            {"min_ttl": None, "meta": None},
            {"ckey": "test.lease", "min_ttl": 300, "meta": ["foo"], "check_server": None},
        ),
        (
            {"min_ttl": 300, "meta": {"foo": True}},
            {"min_ttl": None, "meta": None},
            {"ckey": "test.lease", "min_ttl": 300, "meta": {"foo": True}, "check_server": None},
        ),
        (
            {"min_ttl": 300, "check_server": False},
            {"min_ttl": None, "meta": None},
            {"ckey": "test.lease", "min_ttl": 300, "meta": None, "check_server": False},
        ),
        (
            {"min_ttl": 300, "check_server": True},
            {"min_ttl": None, "meta": None},
            {"ckey": "test.lease", "min_ttl": 300, "meta": None, "check_server": True},
        ),
        (
            {"min_ttl": 300, "check_server": True, "meta": "foo.bar"},
            {"expires_in": -1, "expired": True},
            {
                "ckey": "test.lease",
                "min_ttl": 300,
                "meta": "foo.bar",
                "check_server": True,
                "expires_in": -1,
                "expired": True,
            },
        ),
    ),
)
def test_enrich_info(cfg, info, exp):
    exp["tag"] = "expire"
    res = lease._enrich_info("test.lease", cfg, info)
    assert res == exp



"""
:codeauthor: Pedro Algarvio (pedro@algarvio.me)

.. _runtime_vars:

Runtime Variables
-----------------

:command:`salt-runtests` provides a variable, :py:attr:`RUNTIME_VARS` which has some common paths defined at
startup:

.. autoattribute:: tests.support.runtests.RUNTIME_VARS
    :annotation:

    :TMP: Tests suite temporary directory
    :TMP_CONF_DIR: Configuration directory from where the daemons that :command:`salt-runtests` starts get their
                   configuration files.
    :TMP_CONF_MASTER_INCLUDES: Salt Master configuration files includes directory. See
                               :salt_conf_master:`default_include`.
    :TMP_CONF_MINION_INCLUDES: Salt Minion configuration files includes directory. Seei
                               :salt_conf_minion:`include`.
    :TMP_CONF_CLOUD_INCLUDES: Salt cloud configuration files includes directory. The same as the salt master and
                              minion includes configuration, though under a different directory name.
    :TMP_CONF_CLOUD_PROFILE_INCLUDES: Salt cloud profiles configuration files includes directory. Same as above.
    :TMP_CONF_CLOUD_PROVIDER_INCLUDES: Salt cloud providers configuration files includes directory. Same as above.
    :TMP_SCRIPT_DIR: Temporary scripts directory from where the Salt CLI tools will be called when running tests.
    :TMP_SALT_INTEGRATION_FILES: Temporary directory from where Salt's test suite integration files are copied to.
    :TMP_BASEENV_STATE_TREE: Salt master's **base** environment state tree directory
    :TMP_PRODENV_STATE_TREE: Salt master's **production** environment state tree directory
    :TMP_BASEENV_PILLAR_TREE: Salt master's **base** environment pillar tree directory
    :TMP_PRODENV_PILLAR_TREE: Salt master's **production** environment pillar tree directory


Use it on your test case in case of need. As simple as:

.. code-block:: python

    import os
    from tests.support.runtests import RUNTIME_VARS

    # Path to the testing minion configuration file
    minion_config_path = os.path.join(RUNTIME_VARS.TMP_CONF_DIR, 'minion')

.. _`pytest`: http://pytest.org
"""

import logging
import os
import shutil

import salt.utils.path
import salt.utils.platform

from tests.support import paths  # pylint: disable=import-error

try:
    import pwd
except ImportError:
    import salt.utils.win_functions

log = logging.getLogger(__name__)


def this_user():
    """
    Get the user associated with the current process.
    """
    if salt.utils.platform.is_windows():
        return salt.utils.win_functions.get_current_user(with_domain=False)
    return pwd.getpwuid(os.getuid())[0]


class RootsDict(dict):
    def merge(self, data):
        for key, values in data.items():
            if key not in self:
                self[key] = values
                continue
            for value in values:
                if value not in self[key]:
                    self[key].append(value)
        return self

    def to_dict(self):
        return dict(self)


def recursive_copytree(source, destination, overwrite=False):
    for root, dirs, files in os.walk(source):
        for item in dirs:
            src_path = os.path.join(root, item)
            dst_path = os.path.join(destination, src_path.replace(source, "").lstrip(os.sep))
            if not os.path.exists(dst_path):
                log.debug("Creating directory: %s", dst_path)
                os.makedirs(dst_path)
        for item in files:
            src_path = os.path.join(root, item)
            dst_path = os.path.join(destination, src_path.replace(source, "").lstrip(os.sep))
            if os.path.exists(dst_path) and not overwrite:
                if os.stat(src_path).st_mtime > os.stat(dst_path).st_mtime:
                    log.debug("Copying %s to %s", src_path, dst_path)
                    shutil.copy2(src_path, dst_path)
            else:
                if not os.path.isdir(os.path.dirname(dst_path)):
                    log.debug("Creating directory: %s", os.path.dirname(dst_path))
                    os.makedirs(os.path.dirname(dst_path))
                log.debug("Copying %s to %s", src_path, dst_path)
                shutil.copy2(src_path, dst_path)


class RuntimeVars:

    __self_attributes__ = ("_vars", "_locked", "lock")

    def __init__(self, **kwargs):
        self._vars = kwargs
        self._locked = False

    def lock(self):
        # Late import
        from salt.utils.immutabletypes import freeze  # pylint: disable=import-outside-toplevel

        frozen_vars = freeze(self._vars.copy())
        self._vars = frozen_vars
        self._locked = True

    def __iter__(self):
        yield from self._vars.items()

    def __getattribute__(self, name):
        if name in object.__getattribute__(self, "_vars"):
            return object.__getattribute__(self, "_vars")[name]
        return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        if getattr(self, "_locked", False) is True:
            raise RuntimeError(
                f"After {self.__class__.__name__} is locked, no additional data can be added to it"
            )
        if name in object.__getattribute__(self, "__self_attributes__"):
            object.__setattr__(self, name, value)
            return
        self._vars[name] = value


# <---- Helper Methods -----------------------------------------------------------------------------------------------


# ----- Global Variables -------------------------------------------------------------------------------------------->
XML_OUTPUT_DIR = os.environ.get(
    "SALT_XML_TEST_REPORTS_DIR", os.path.join(paths.TMP, "xml-test-reports")
)
# <---- Global Variables ---------------------------------------------------------------------------------------------


# ----- Tests Runtime Variables ------------------------------------------------------------------------------------->

RUNTIME_VARS = RuntimeVars(
    TMP=paths.TMP,
    SYS_TMP_DIR=paths.SYS_TMP_DIR,
    FILES=paths.FILES,
    CONF_DIR=paths.CONF_DIR,
    PILLAR_DIR=paths.PILLAR_DIR,
    ENGINES_DIR=paths.ENGINES_DIR,
    LOG_HANDLERS_DIR=paths.LOG_HANDLERS_DIR,
    TMP_ROOT_DIR=paths.TMP_ROOT_DIR,
    TMP_CONF_DIR=paths.TMP_CONF_DIR,
    TMP_MINION_CONF_DIR=paths.TMP_MINION_CONF_DIR,
    TMP_CONF_MASTER_INCLUDES=os.path.join(paths.TMP_CONF_DIR, "master.d"),
    TMP_CONF_MINION_INCLUDES=os.path.join(paths.TMP_CONF_DIR, "minion.d"),
    TMP_CONF_PROXY_INCLUDES=os.path.join(paths.TMP_CONF_DIR, "proxy.d"),
    TMP_CONF_CLOUD_INCLUDES=os.path.join(paths.TMP_CONF_DIR, "cloud.conf.d"),
    TMP_CONF_CLOUD_PROFILE_INCLUDES=os.path.join(paths.TMP_CONF_DIR, "cloud.profiles.d"),
    TMP_CONF_CLOUD_PROVIDER_INCLUDES=os.path.join(paths.TMP_CONF_DIR, "cloud.providers.d"),
    TMP_SUB_MINION_CONF_DIR=paths.TMP_SUB_MINION_CONF_DIR,
    TMP_SYNDIC_MASTER_CONF_DIR=paths.TMP_SYNDIC_MASTER_CONF_DIR,
    TMP_SYNDIC_MINION_CONF_DIR=paths.TMP_SYNDIC_MINION_CONF_DIR,
    TMP_SSH_CONF_DIR=paths.TMP_SSH_CONF_DIR,
    TMP_SCRIPT_DIR=paths.TMP_SCRIPT_DIR,
    TMP_STATE_TREE=paths.TMP_STATE_TREE,
    TMP_BASEENV_STATE_TREE=paths.TMP_STATE_TREE,
    TMP_PILLAR_TREE=paths.TMP_PILLAR_TREE,
    TMP_BASEENV_PILLAR_TREE=paths.TMP_PILLAR_TREE,
    TMP_PRODENV_STATE_TREE=paths.TMP_PRODENV_STATE_TREE,
    TMP_PRODENV_PILLAR_TREE=paths.TMP_PRODENV_PILLAR_TREE,
    SHELL_TRUE_PATH=(
        salt.utils.path.which("true")
        if not salt.utils.platform.is_windows()
        else "cmd /c exit 0 > nul"
    ),
    SHELL_FALSE_PATH=(
        salt.utils.path.which("false")
        if not salt.utils.platform.is_windows()
        else "cmd /c exit 1 > nul"
    ),
    RUNNING_TESTS_USER=this_user(),
    RUNTIME_CONFIGS={},
    CODE_DIR=paths.CODE_DIR,
    SALT_CODE_DIR=paths.SALT_CODE_DIR,
    BASE_FILES=paths.BASE_FILES,
    PROD_FILES=paths.PROD_FILES,
    TESTS_DIR=paths.TESTS_DIR,
)
# <---- Tests Runtime Variables --------------------------------------------------------------------------------------



import tempfile
from pathlib import Path

from . import prompt
from .cmd import CommandNotFound
from .cmd import local
from .copier import discover_project_name

# Should follow the version used for relenv packages, see
# https://github.com/saltstack/salt/blob/master/cicd/shared-gh-workflows-context.yml
RECOMMENDED_PYVER = "3.10"
# For discovery of existing virtual environment, descending priority.
VENV_DIRS = (
    ".venv",
    "venv",
    ".env",
    "env",
)


try:
    uv = local["uv"]
except CommandNotFound:
    uv = None


def is_venv(path):
    if (venv_path := Path(path)).is_dir and (venv_path / "pyvenv.cfg").exists():
        return venv_path
    return False


def discover_venv(project_root="."):
    base = Path(project_root).resolve()
    for name in VENV_DIRS:
        if found := is_venv(base / name):
            return found
    raise RuntimeError(f"No venv found in {base}")


def create_venv(project_root=".", directory=None):
    base = Path(project_root).resolve()
    venv = (base / (directory or VENV_DIRS[0])).resolve()
    if is_venv(venv):
        raise RuntimeError(f"Venv at {venv} already exists")
    prompt.status(f"Creating virtual environment at {venv}")
    if uv is not None:
        prompt.status("Found `uv`. Creating venv")
        uv(
            "venv",
            # Install pip/setuptools/wheel for compatibility
            "--seed",
            "--python",
            RECOMMENDED_PYVER,
            f"--prompt=saltext-{discover_project_name()}",
        )
    else:
        prompt.status("Did not find `uv`. Falling back to `venv`")
        try:
            python = local[f"python{RECOMMENDED_PYVER}"]
        except CommandNotFound:
            python = local["python3"]
            version = python("--version").split(" ")[1]
            if not version.startswith(RECOMMENDED_PYVER):
                raise RuntimeError(
                    f"No `python{RECOMMENDED_PYVER}` executable found in $PATH, exiting"
                )
        python("-m", "venv", VENV_DIRS[0], f"--prompt=saltext-{discover_project_name()}")
    return venv


def ensure_project_venv(project_root=".", reinstall=True, install_extras=False):
    exists = False
    try:
        venv = discover_venv(project_root)
        prompt.status(f"Found existing virtual environment at {venv}")
        exists = True
    except RuntimeError:
        venv = create_venv(project_root)
    if not reinstall:
        return venv
    extras = ["dev", "tests", "docs"]
    if install_extras:
        extras.append("dev_extra")
    prompt.status(("Reinstalling" if exists else "Installing") + " project and dependencies")
    with local.venv(venv):
        if uv is not None:
            uv("pip", "install", "-e", f".[{','.join(extras)}]")
        else:
            try:
                # We install uv into the virtualenv, so it might be available now.
                # It speeds up this step a lot.
                local["uv"]("pip", "install", "-e", f".[{','.join(extras)}]")
            except CommandNotFound:
                # Salt does not build correctly with setuptools >= 75.6.0.
                # uv reads this constraint from pyproject.toml, but pip needs this workaround.
                with tempfile.NamedTemporaryFile(delete=False) as constraints_file:
                    setuptools_constraint = "setuptools<75.6.0"
                    constraints_file.write(setuptools_constraint.encode())
                try:
                    with local.env(PIP_CONSTRAINT=constraints_file.name):
                        local["python"]("-m", "pip", "install", "-e", f".[{','.join(extras)}]")
                finally:
                    Path(constraints_file.name).unlink()
        if not exists or not (Path(project_root) / ".git" / "hooks" / "pre-commit").exists():
            prompt.status("Installing pre-commit hooks")
            local["python"]("-m", "pre_commit", "install", "--install-hooks")
    return venv



import logging
from unittest.mock import ANY
from unittest.mock import patch

import pytest
import salt.exceptions

import saltext.vault.utils.vault as vaultutil
from saltext.vault.modules import vault


@pytest.fixture
def configure_loader_modules():
    return {
        vault: {
            "__grains__": {"id": "test-minion"},
        }
    }


@pytest.fixture
def data():
    return {"foo": "bar"}


@pytest.fixture
def policy_response():
    return {
        "name": "test-policy",
        "rules": 'path "secret/*"\\n{\\n  capabilities = ["read"]\\n}',
    }


@pytest.fixture
def policies_list_response():
    return {
        "policies": ["default", "root", "test-policy"],
    }


@pytest.fixture
def data_list():
    return ["foo"]


@pytest.fixture
def read_kv(data):
    with patch("saltext.vault.utils.vault.read_kv", autospec=True) as read:
        read.return_value = data
        yield read


@pytest.fixture
def list_kv(data_list):
    with patch("saltext.vault.utils.vault.list_kv", autospec=True) as _list:
        _list.return_value = data_list
        yield _list


@pytest.fixture
def read_kv_not_found(read_kv):
    read_kv.side_effect = vaultutil.VaultNotFoundError
    yield read_kv


@pytest.fixture
def list_kv_not_found(list_kv):
    list_kv.side_effect = vaultutil.VaultNotFoundError
    yield list_kv


@pytest.fixture
def write_kv():
    with patch("saltext.vault.utils.vault.write_kv", autospec=True) as write:
        yield write


@pytest.fixture
def write_kv_err(write_kv):
    write_kv.side_effect = vaultutil.VaultPermissionDeniedError("damn")
    yield write_kv


@pytest.fixture
def patch_kv():
    with patch("saltext.vault.utils.vault.patch_kv", autospec=True) as patch_kv:
        yield patch_kv


@pytest.fixture
def patch_kv_err(patch_kv):
    patch_kv.side_effect = vaultutil.VaultPermissionDeniedError("damn")
    yield patch_kv


@pytest.fixture
def delete_kv():
    with patch("saltext.vault.utils.vault.delete_kv", autospec=True) as delete_kv:
        yield delete_kv


@pytest.fixture
def delete_kv_err(delete_kv):
    delete_kv.side_effect = vaultutil.VaultPermissionDeniedError("damn")
    yield delete_kv


@pytest.fixture
def destroy_kv():
    with patch("saltext.vault.utils.vault.destroy_kv", autospec=True) as destroy_kv:
        yield destroy_kv


@pytest.fixture
def destroy_kv_err(destroy_kv):
    destroy_kv.side_effect = vaultutil.VaultPermissionDeniedError("damn")
    yield destroy_kv


@pytest.fixture
def query():
    with patch("saltext.vault.utils.vault.query", autospec=True) as query:
        yield query


@pytest.mark.usefixtures("read_kv")
@pytest.mark.parametrize("key,expected", [(None, {"foo": "bar"}), ("foo", "bar")])
def test_read_secret(key, expected):
    """
    Ensure read_secret works as expected without and with specified key.
    KV v1/2 is handled in the utils module.
    """
    res = vault.read_secret("some/path", key=key)
    assert res == expected


@pytest.mark.usefixtures("read_kv_not_found", "list_kv_not_found")
@pytest.mark.parametrize("func", ["read_secret", "list_secrets"])
def test_read_list_secret_with_default(func):
    """
    Ensure read_secret and list_secrets with defaults set return those
    if the path was not found.
    """
    tgt = getattr(vault, func)
    res = tgt("some/path", default=["f"])
    assert res == ["f"]


@pytest.mark.usefixtures("read_kv_not_found", "list_kv_not_found")
@pytest.mark.parametrize("func", ["read_secret", "list_secrets"])
def test_read_list_secret_without_default(func):
    """
    Ensure read_secret and list_secrets without defaults set raise
    a CommandExecutionError when the path is not found.
    """
    tgt = getattr(vault, func)
    with pytest.raises(salt.exceptions.CommandExecutionError, match=".*VaultNotFoundError.*"):
        tgt("some/path")


@pytest.mark.usefixtures("list_kv")
@pytest.mark.parametrize(
    "keys_only,expected",
    [
        (False, {"keys": ["foo"]}),
        (True, ["foo"]),
    ],
)
def test_list_secrets(keys_only, expected):
    """
    Ensure list_secrets works as expected. keys_only=False is default to
    stay backwards-compatible. There should not be a reason to have the
    function return a dict with a single predictable key otherwise.
    """
    res = vault.list_secrets("some/path", keys_only=keys_only)
    assert res == expected


def test_write_secret(data, write_kv):
    """
    Ensure write_secret parses kwargs as expected
    """
    path = "secret/some/path"
    res = vault.write_secret(path, **data)
    assert res
    write_kv.assert_called_once_with(path, data, opts=ANY, context=ANY)


@pytest.mark.usefixtures("write_kv_err")
def test_write_secret_err(data, caplog):
    """
    Ensure write_secret handles exceptions as expected
    """
    with caplog.at_level(logging.ERROR):
        res = vault.write_secret("secret/some/path", **data)
        assert not res
        assert "Failed to write secret! VaultPermissionDeniedError: damn" in caplog.messages


def test_write_raw(data, write_kv):
    """
    Ensure write_secret works as expected
    """
    path = "secret/some/path"
    res = vault.write_raw(path, data)
    assert res
    write_kv.assert_called_once_with(path, data, opts=ANY, context=ANY)


@pytest.mark.usefixtures("write_kv_err")
def test_write_raw_err(data, caplog):
    """
    Ensure write_raw handles exceptions as expected
    """
    with caplog.at_level(logging.ERROR):
        res = vault.write_raw("secret/some/path", data)
        assert not res
        assert "Failed to write secret! VaultPermissionDeniedError: damn" in caplog.messages


def test_patch_secret(data, patch_kv):
    """
    Ensure patch_secret parses kwargs as expected
    """
    path = "secret/some/path"
    res = vault.patch_secret(path, **data)
    assert res
    patch_kv.assert_called_once_with(path, data, opts=ANY, context=ANY)


@pytest.mark.usefixtures("patch_kv_err")
def test_patch_secret_err(data, caplog):
    """
    Ensure patch_secret handles exceptions as expected
    """
    with caplog.at_level(logging.ERROR):
        res = vault.patch_secret("secret/some/path", **data)
        assert not res
        assert "Failed to patch secret! VaultPermissionDeniedError: damn" in caplog.messages


@pytest.mark.parametrize("args", [[], [1, 2]])
def test_delete_secret(delete_kv, args):
    """
    Ensure delete_secret works as expected
    """
    path = "secret/some/path"
    res = vault.delete_secret(path, *args)
    assert res
    delete_kv.assert_called_once_with(
        path, opts=ANY, context=ANY, versions=args or None, all_versions=False
    )


@pytest.mark.usefixtures("delete_kv_err")
@pytest.mark.parametrize("args", [[], [1, 2]])
def test_delete_secret_err(args, caplog):
    """
    Ensure delete_secret handles exceptions as expected
    """
    with caplog.at_level(logging.ERROR):
        res = vault.delete_secret("secret/some/path", *args)
        assert not res
        assert "Failed to delete secret! VaultPermissionDeniedError: damn" in caplog.messages


@pytest.mark.parametrize("args", [[], [1], [1, 2]])
def test_destroy_secret(destroy_kv, args):
    """
    Ensure destroy_secret works as expected
    """
    path = "secret/some/path"
    res = vault.destroy_secret(path, *args)
    assert res
    destroy_kv.assert_called_once_with(
        path, args or None, opts=ANY, context=ANY, all_versions=False
    )


@pytest.mark.usefixtures("destroy_kv_err")
@pytest.mark.parametrize("args", [[1], [1, 2]])
def test_destroy_secret_err(caplog, args):
    """
    Ensure destroy_secret handles exceptions as expected
    """
    with caplog.at_level(logging.ERROR):
        res = vault.destroy_secret("secret/some/path", *args)
        assert not res
        assert "Failed to destroy secret! VaultPermissionDeniedError: damn" in caplog.messages


def test_clear_token_cache():
    """
    Ensure clear_token_cache wraps the utility function properly
    """
    with patch("saltext.vault.utils.vault.clear_cache") as cache:
        vault.clear_token_cache()
        cache.assert_called_once_with(ANY, ANY, connection=True, session=False)


def test_policy_fetch(query, policy_response):
    """
    Ensure policy_fetch returns rules only and calls the API as expected
    """
    query.return_value = policy_response
    res = vault.policy_fetch("test-policy")
    assert res == policy_response["rules"]
    query.assert_called_once_with("GET", "sys/policy/test-policy", opts=ANY, context=ANY)


def test_policy_fetch_not_found(query):
    """
    Ensure policy_fetch returns None when the policy was not found
    """
    query.side_effect = vaultutil.VaultNotFoundError
    res = vault.policy_fetch("test-policy")
    assert res is None


@pytest.mark.parametrize(
    "func,args",
    [
        ("policy_fetch", []),
        ("policy_write", ["rule"]),
        ("policy_delete", []),
        ("policies_list", None),
    ],
)
def test_policy_functions_raise_errors(query, func, args):
    """
    Ensure policy functions raise CommandExecutionErrors
    """
    query.side_effect = vaultutil.VaultPermissionDeniedError
    func = getattr(vault, func)
    with pytest.raises(
        salt.exceptions.CommandExecutionError, match=".*VaultPermissionDeniedError.*"
    ):
        if args is None:
            func()
        else:
            func("test-policy", *args)


def test_policy_write(query, policy_response):
    """
    Ensure policy_write calls the API as expected
    """
    query.return_value = True
    res = vault.policy_write("test-policy", policy_response["rules"])
    assert res
    query.assert_called_once_with(
        "POST",
        "sys/policy/test-policy",
        opts=ANY,
        context=ANY,
        payload={"policy": policy_response["rules"]},
        safe_to_retry=True,
    )


def test_policy_delete(query):
    """
    Ensure policy_delete calls the API as expected
    """
    query.return_value = True
    res = vault.policy_delete("test-policy")
    assert res
    query.assert_called_once_with("DELETE", "sys/policy/test-policy", opts=ANY, context=ANY)


def test_policy_delete_handles_not_found(query):
    """
    Ensure policy_delete returns False instead of raising CommandExecutionError
    when a policy was absent already.
    """
    query.side_effect = vaultutil.VaultNotFoundError
    res = vault.policy_delete("test-policy")
    assert not res


def test_policies_list(query, policies_list_response):
    """
    Ensure policies_list returns policy list only and calls the API as expected
    """
    query.return_value = policies_list_response
    res = vault.policies_list()
    assert res == policies_list_response["policies"]
    query.assert_called_once_with("GET", "sys/policy", opts=ANY, context=ANY)


@pytest.mark.parametrize("method", ["POST", "DELETE"])
@pytest.mark.parametrize("payload", [None, {"data": {"foo": "bar"}}])
def test_query(query, method, payload):
    """
    Ensure query wraps the utility function properly
    """
    query.return_value = True
    endpoint = "test/endpoint"
    res = vault.query(method, endpoint, payload=payload)
    assert res
    query.assert_called_once_with(method, endpoint, opts=ANY, context=ANY, payload=payload)


def test_query_raises_errors(query):
    """
    Ensure query raises CommandExecutionErrors
    """
    query.side_effect = vaultutil.VaultPermissionDeniedError
    with pytest.raises(
        salt.exceptions.CommandExecutionError, match=".*VaultPermissionDeniedError.*"
    ):
        vault.query("GET", "test/endpoint")



"""
SSH wrapper for the :py:mod:`vault <saltext.vault.modules.vault>` execution module.

See there for documentation.
"""

from salt.utils.functools import namespaced_function

from saltext.vault.modules.vault import clear_cache
from saltext.vault.modules.vault import clear_token_cache
from saltext.vault.modules.vault import delete_secret
from saltext.vault.modules.vault import destroy_secret
from saltext.vault.modules.vault import get_server_config
from saltext.vault.modules.vault import list_secrets
from saltext.vault.modules.vault import patch_secret
from saltext.vault.modules.vault import policies_list
from saltext.vault.modules.vault import policy_delete
from saltext.vault.modules.vault import policy_fetch
from saltext.vault.modules.vault import policy_write
from saltext.vault.modules.vault import query
from saltext.vault.modules.vault import read_secret
from saltext.vault.modules.vault import read_secret_meta
from saltext.vault.modules.vault import restore_secret
from saltext.vault.modules.vault import update_config
from saltext.vault.modules.vault import wipe_secret
from saltext.vault.modules.vault import write_raw
from saltext.vault.modules.vault import write_secret

clear_cache = namespaced_function(clear_cache, globals())
clear_token_cache = namespaced_function(clear_token_cache, globals())
delete_secret = namespaced_function(delete_secret, globals())
destroy_secret = namespaced_function(destroy_secret, globals())
get_server_config = namespaced_function(get_server_config, globals())
list_secrets = namespaced_function(list_secrets, globals())
patch_secret = namespaced_function(patch_secret, globals())
policies_list = namespaced_function(policies_list, globals())
policy_delete = namespaced_function(policy_delete, globals())
policy_fetch = namespaced_function(policy_fetch, globals())
policy_write = namespaced_function(policy_write, globals())
query = namespaced_function(query, globals())
read_secret = namespaced_function(read_secret, globals())
read_secret_meta = namespaced_function(read_secret_meta, globals())
restore_secret = namespaced_function(restore_secret, globals())
update_config = namespaced_function(update_config, globals())
wipe_secret = namespaced_function(wipe_secret, globals())
write_raw = namespaced_function(write_raw, globals())
write_secret = namespaced_function(write_secret, globals())



"""
Unit tests for the Vault runner

This module only tests a deprecated function, see
tests/unit/runners/test_vault.py for the current tests.
"""

import logging
from unittest.mock import ANY
from unittest.mock import Mock
from unittest.mock import patch

import pytest

import saltext.vault.utils.vault as vaultutil
from saltext.vault.runners import vault
from saltext.vault.utils.vault import client as vclient

pytestmark = [
    pytest.mark.usefixtures("validate_sig", "policies"),
]

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {
        vault: {
            "__opts__": {
                "vault": {
                    "url": "http://127.0.0.1",
                    "auth": {
                        "token": "test",
                        "method": "token",
                        "allow_minion_override": True,
                    },
                }
            }
        }
    }


@pytest.fixture
def auth():
    return {
        "auth": {
            "client_token": "test",
            "renewable": False,
            "lease_duration": 0,
        }
    }


@pytest.fixture
def client(auth):
    client_mock = Mock(vclient.AuthenticatedVaultClient)
    client_mock.post.return_value = auth
    with patch("saltext.vault.runners.vault._get_master_client", Mock(return_value=client_mock)):
        yield client_mock


@pytest.fixture
def validate_sig():
    with patch("saltext.vault.runners.vault._validate_signature", autospec=True, return_value=None):
        yield


@pytest.fixture
def policies():
    with patch("saltext.vault.runners.vault._get_policies_cached", autospec=True) as policies:
        policies.return_value = ["saltstack/minion/test-minion", "saltstack/minions"]
        yield policies


# Basic tests for test_generate_token: all exits


def test_generate_token(client):
    result = vault.generate_token("test-minion", "signature")
    log.debug("generate_token result: %s", result)
    assert isinstance(result, dict)
    assert "error" not in result
    assert "token" in result
    assert result["token"] == "test"
    client.post.assert_called_with("auth/token/create", payload=ANY, wrap=False)


def test_generate_token_uses(client):
    # Test uses
    num_uses = 6
    result = vault.generate_token("test-minion", "signature", uses=num_uses)
    assert "uses" in result
    assert result["uses"] == num_uses
    json_request = {
        "policies": ["saltstack/minion/test-minion", "saltstack/minions"],
        "num_uses": num_uses,
        "meta": {
            "saltstack-jid": "<no jid set>",
            "saltstack-minion": "test-minion",
            "saltstack-user": "<no user set>",
        },
    }
    client.post.assert_called_with("auth/token/create", payload=json_request, wrap=False)


def test_generate_token_ttl(client):
    # Test ttl
    expected_ttl = "6h"
    result = vault.generate_token("test-minion", "signature", ttl=expected_ttl)
    assert result["uses"] == 1
    json_request = {
        "policies": ["saltstack/minion/test-minion", "saltstack/minions"],
        "num_uses": 1,
        "explicit_max_ttl": expected_ttl,
        "meta": {
            "saltstack-jid": "<no jid set>",
            "saltstack-minion": "test-minion",
            "saltstack-user": "<no user set>",
        },
    }
    client.post.assert_called_with("auth/token/create", payload=json_request, wrap=False)


def test_generate_token_permission_denied(client):
    client.post.side_effect = vaultutil.VaultPermissionDeniedError("no reason")
    result = vault.generate_token("test-minion", "signature")
    assert isinstance(result, dict)
    assert "error" in result
    assert result["error"] == "VaultPermissionDeniedError: no reason"


def test_generate_token_exception(client):
    client.post.side_effect = Exception("Test Exception Reason")
    result = vault.generate_token("test-minion", "signature")
    assert isinstance(result, dict)
    assert "error" in result
    assert result["error"] == "Exception: Test Exception Reason"


def test_generate_token_no_matching_policies(policies):
    policies.return_value = []
    result = vault.generate_token("test-minion", "signature")
    assert isinstance(result, dict)
    assert "error" in result
    assert result["error"] == "SaltRunnerError: No policies matched minion."



import pytest


@pytest.fixture
def beacons(loaders):
    def _beacons(self):
        """
        The beacons loaded by the salt loader.
        We need to patch this in since it's currently not
        supported by pytest-salt-factories.
        """
        # Do not move these deferred imports. It allows running against a Salt
        # onedir build in salt's repo checkout.
        import salt.loader  # pylint: disable=import-outside-toplevel

        if self._beacons is None:
            self._beacons = salt.loader.beacons(
                self.opts,
                functions=self.modules,
                context=self.context,
                loaded_base_name=self.loaded_base_name,
            )
        return self._beacons

    def _reload_all(self):
        loaders.reload_all()
        if self._beacons is not None:
            self._beacons.clean_modules()
            self._beacons.clear()
            self._beacons = None

    try:
        loaders._beacons = None
        type(loaders).beacons = property(_beacons)
        loaders.reload_all = _reload_all
        yield loaders.beacons
    finally:
        loaders.reset_state()
        delattr(loaders, "_beacons")
        delattr(type(loaders), "beacons")



"""
Manage the Vault (or OpenBao) SSH secret engine.

.. versionadded:: 1.2.0

.. important::
    This module requires the general :ref:`Vault setup <vault-setup>`.
"""

import logging

import salt.utils.dictdiffer
from salt.exceptions import CommandExecutionError
from salt.exceptions import SaltInvocationError

from saltext.vault.utils.vault.helpers import deserialize_csl
from saltext.vault.utils.vault.helpers import timestring_map

log = logging.getLogger(__name__)

LIST_ROLE_PARAMS = (
    "allowed_users",
    "allowed_critical_options",
    "allowed_domains",
    "allowed_extensions",
    "cidr_list",
    "exclude_cidr_list",
)
MAP_ROLE_PARAMS = ("default_critical_options", "default_extensions", "allowed_user_key_lengths")
TIME_ROLE_PARAMS = ("ttl", "max_ttl", "not_before_duration")


def ca_present(
    name,
    private_key=None,
    public_key=None,
    key_type="ssh-rsa",
    key_bits=0,
    mount="ssh",
):
    """
    Ensure a CA is present on the mount. Note that only one is possible
    per mount. This state will not inspect the properties once a CA
    has been initialized.

    name
        Irrelevant.

    private_key
        The private key part of the SSH CA key pair. Can be a file
        local to the minion or a PEM-encoded string.
        If this or ``public_key`` is unspecified, will generate a pair
        on the Vault server.

    public_key
        The public key part of the SSH CA key pair. Can be a file
        local to the minion or a PEM-encoded string.
        If this or ``public_key`` is unspecified, will generate a pair
        on the Vault server.

    key_type
        The desired key type for the generated SSH CA key when generating
        on the Vault server. Valid: ``ssh-rsa`` (default), ``sha2-nistp256``,
        ``ecdsa-sha2-nistp384``, ``ecdsa-sha2-nistp521``, or ``ssh-ed25519``.
        Can also specify an algorithm: ``rsa``, ``ec``, or ``ed25519``.

    key_bits
        The desired key bits for the generated SSH CA key when generating
        on the Vault server. Only used for variable length keys (e.g. ``ssh-rsa``)
        or when ``ec`` was specified as ``key_type``, in which case this
        selects the NIST P-curve: ``256``, ``384``, ``521``.
        0 (default) will select 4096 bits for RSA or NIST P-256 for EC.

    mount
        The name of the mount point the SSH secret backend is mounted at.
        Defaults to ``ssh``.
    """
    ret = {
        "name": name,
        "result": True,
        "comment": "The CA has already been initialized",
        "changes": {},
    }

    try:
        try:
            current = __salt__["vault_ssh.read_ca"](mount=mount)
        except CommandExecutionError as err:
            if (
                "keys haven't been configured yet" not in err.message  # Vault
                and "no default issuer currently configured" not in err.message  # OpenBao
            ):
                raise
            current = None
        if current:
            return ret

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "The CA would have been initialized"
            ret["changes"]["created"] = f"SSH CA on mount {mount}"
            return ret

        ca = __salt__["vault_ssh.create_ca"](
            private_key=private_key,
            public_key=public_key,
            key_type=key_type,
            key_bits=key_bits,
            mount=mount,
        )
        ret["comment"] = "The CA has been initialized"
        ret["changes"]["created"] = ca

    except (CommandExecutionError, SaltInvocationError) as err:
        ret["result"] = False
        ret["comment"] = str(err)
        ret["changes"] = {}

    return ret


def ca_absent(
    name,
    mount="ssh",
):
    """
    Ensure a CA is absent from the mount. Note that you will not be
    able to easily recover a destroy private key.

    name
        Irrelevant.

    mount
        The name of the mount point the SSH secret backend is mounted at.
        Defaults to ``ssh``.
    """
    ret = {
        "name": name,
        "result": True,
        "comment": "There is no CA on the mount",
        "changes": {},
    }

    try:
        try:
            current = __salt__["vault_ssh.read_ca"](mount=mount)
        except CommandExecutionError as err:
            if (
                "keys haven't been configured yet" not in err.message  # Vault
                and "no default issuer currently configured" not in err.message  # OpenBao
            ):
                raise
            current = None
        if not current:
            return ret

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "The CA would have been destroyed"
            ret["changes"]["destroyed"] = f"SSH CA on mount {mount}"
            return ret

        ca = __salt__["vault_ssh.destroy_ca"](mount=mount)
        ret["comment"] = "The CA has been destroyed"
        ret["changes"]["destroyed"] = ca

    except (CommandExecutionError, SaltInvocationError) as err:
        ret["result"] = False
        ret["comment"] = str(err)
        ret["changes"] = {}

    return ret


def role_present_otp(
    name,
    default_user,
    cidr_list=None,
    allowed_users=None,
    exclude_cidr_list=None,
    port=None,
    mount="ssh",
):
    """
    Ensure an SSH role (OTP type) is present as specified.

    name
        The name of the SSH role.

    default_user
        The default username for which a credential will be generated.
        Required.

    cidr_list
        List of CIDR blocks to which the role is applicable.
        Required, unless the role is registered as a zero-address role.

    allowed_users
        List of usernames the client can request under this role.
        By default, **any usernames are allowed** (``*``).
        The ``default_user`` will always be allowed.

    exclude_cidr_list
        List of CIDR blocks not accepted by the role.

    port
        Specifies the port number for SSH connections, which will be returned to
        OTP clients as an informative value. Defaults to ``22``.

    mount
        The name of the mount point the SSH secret backend is mounted at.
        Defaults to ``ssh``.
    """
    ret = {
        "name": name,
        "result": True,
        "comment": "The role is present as specified",
        "changes": {},
    }

    try:
        port = int(port) if port else None
    except TypeError:
        ret["result"] = False
        ret["comment"] = "'port' must be castable to an integer"
        return ret

    payload = {
        "default_user": default_user,
        "cidr_list": cidr_list,
        "allowed_users": allowed_users,
        "exclude_cidr_list": exclude_cidr_list,
        "port": port,
    }
    return _role_present(name, "otp", ret, mount=mount, **payload)


def _diff_role_params(curr, wanted):
    diff = {}
    for param, val in wanted.items():
        if param in LIST_ROLE_PARAMS:
            curr_param = set(deserialize_csl(curr.get(param, [])))
            wanted_param = set(deserialize_csl(val or []))
            added = wanted_param - curr_param
            removed = curr_param - wanted_param
            if added or removed:
                diff[param] = {"added": list(sorted(added)), "removed": list(sorted(removed))}
            continue
        if param in MAP_ROLE_PARAMS:
            val = val or {}
            if param == "allowed_user_key_lengths":
                for algo, allowed in val.items():
                    if isinstance(allowed, int):
                        val[algo] = [allowed]
                    else:
                        val[algo] = deserialize_csl(allowed)
            map_diff = salt.utils.dictdiffer.recursive_diff(
                curr.get(param, {}), val, ignore_missing_keys=False
            )
            if map_diff.added() or map_diff.changed() or map_diff.removed():
                diff[param] = {
                    "added": map_diff.added(),
                    "changed": map_diff.changed(),
                    "removed": map_diff.removed(),
                }
            continue
        curr_val = curr.get(param)
        if param in TIME_ROLE_PARAMS:
            curr_val = timestring_map(curr_val)
            val = timestring_map(val)
        if curr_val != val:
            diff[param] = {"old": curr_val, "new": val}
    return diff


def _role_present(name, key_type, ret, mount="ssh", **kwargs):
    try:
        try:
            current = __salt__["vault_ssh.read_role"](name, mount=mount)
        except CommandExecutionError as err:
            if "VaultNotFoundError" not in str(err):
                raise
            current = None
            ret["changes"]["created"] = name
        if current:
            if current["key_type"] != key_type:
                ret["changes"]["key_type"] = {"old": current["key_type"], "new": key_type}
            else:
                ret["changes"] = _diff_role_params(current, kwargs)

        if not ret["changes"]:
            return ret

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Role `{name}` would have been {'updated' if current else 'created'}"
            return ret

        __salt__[f"vault_ssh.write_role_{key_type}"](
            name,
            **kwargs,
            mount=mount,
        )

        try:
            new = __salt__["vault_ssh.read_role"](name, mount=mount)
        except CommandExecutionError as err:
            if "VaultNotFoundError" not in str(err):
                raise
            raise CommandExecutionError(
                "There were no errors during role management, but it is reported as absent."
            ) from err

        new_diff = _diff_role_params(new, kwargs)

        if new_diff:
            ret["result"] = False
            ret["comment"] = (
                "There were no errors during role management, but "
                f"the reported parameters do not match: {new_diff}"
            )
        else:
            ret["comment"] = f"Role `{name}` has been {'updated' if current else 'created'}"

    except (CommandExecutionError, SaltInvocationError) as err:
        ret["result"] = False
        ret["comment"] = str(err)
        ret["changes"] = {}

    return ret


def role_present_ca(
    name,
    default_user="",
    default_user_template=False,
    allowed_users=None,
    allowed_users_template=False,
    allowed_domains=None,
    allowed_domains_template=False,
    ttl=0,
    max_ttl=0,
    allowed_critical_options=None,
    allowed_extensions=None,
    default_critical_options=None,
    default_extensions=None,
    default_extensions_template=False,
    allow_user_certificates=False,
    allow_host_certificates=False,
    allow_bare_domains=False,
    allow_subdomains=False,
    allow_user_key_ids=False,
    key_id_format="",
    allowed_user_key_lengths=None,
    algorithm_signer="default",
    not_before_duration=30,
    mount="ssh",
):
    """
    Ensure an SSH role (CA type) is present as specified.

    name
        The name of the SSH role.

    default_user
        The default username for which a credential will be generated.
        When ``default_user_template`` is true, this can contain an identity
        template with any prefix or suffix, like ``ssh-{{identity.entity.id}}-user``.
        If you wish this to be a valid principal, it must also be in ``allowed_users``.

    default_user_template
        Allow ``default_users`` to be specified using identity template values.
        A non-templated user is also permitted. Defaults to false.

    allowed_users
        List of usernames the client can request under this role.
        By default, none are allowed. Set this to ``*`` to allow any usernames.
        If ``allowed_users_template`` is true, this list can contain an
        identity template with any prefix or suffix. The ``default_user``
        will always be allowed.

    allowed_users_template
        Allow ``allowed_users`` to be specified using identity template values.
        Non-templated users are also permitted. Defaults to false.

    allowed_domains
        List of domains for which a client can request a host certificate.
        ``*`` allows any domain. See also ``allow_bare_domains`` and ``allow_subdomains``.

    allowed_domains_template
        Allow ``allowed_domains`` to be specified using identity template values.
        Non-templated domains are also permitted. Defaults to false.

    ttl
        Specifies the Time To Live value provided as a string duration with
        time suffix. Hour is the largest suffix. If unset, uses the system
        default value or the value of ``max_ttl``, whichever is shorter

    max_ttl
        Specifies the maximum Time To Live provided as a string duration with
        time suffix. Hour is the largest suffix. If unset, defaults to the
        system maximum lease TTL.

    allowed_critical_options
        List of critical options that certificates can carry when signed.
        If unset (default), allows any option.

    allowed_extensions
        List of extensions that certificates can carry when signed.
        If unset (default), will always take the extensions
        from ``default_extensions`` only. If set to ``*``, will allow
        any extension to be set.
        For the list of extensions, take a look at the sshd manual's
        AUTHORIZED_KEYS FILE FORMAT section. You should add a ``permit-``
        before the name of extension to allow it.

    default_critical_options
        Map of critical options to their values certificates should carry
        if none are provided when signing.

    default_extensions
        Map of extensions to their values certificates should carry
        if none are provided when signing or allowed_extensions is unset.

    default_extensions_template
        Allow ``default_extensions`` to be specified using identity template values.
        Non-templated values are also permitted. Defaults to false.

    allow_user_certificates
        Allow certificates to be signed for ``user`` use. Defaults to false.

    allow_host_certificates
        Allow certificates to be signed for ``host`` use. Defaults to false.

    allow_bare_domains
        Allow host certificates to be signed for the base domains listed in
        ``allowed_domains``. This is a separate option as in some cases this
        can be considered a security threat. Defaults to false.

    allow_subdomains
        Allow host certificates to be signed for subdomains of the base domains
        listed in ``allowed_domains``. Defaults to false.

    allow_user_key_ids
        Allow users to override the key ID for a certificate. When false (default),
        the key ID will always be the token display name.
        The key ID is logged by the SSH server and can be useful for auditing.

    key_id_format
        Specifies a custom format for the key ID of a signed certificate.
        See `key_id_format <https://developer.hashicorp.com/vault/api-docs/secret/ssh#key_id_format>`_
        for available template values.

    allowed_user_key_lengths
        Map of ssh key types to allowed sizes when signing with the CA type.
        Values can be a list of multiple sizes.
        Keys can both be OpenSSH-style key identifiers and short names
        (``rsa``, ``ecdsa``, ``dsa``, or ``ed25519``). If an algorithm has
        a fixed key size, values are ignored.

    algorithm_signer
        **RSA** algorithm to sign keys with. Valid: ``ssh-rsa``, ``rsa-sha2-256``,
        ``rsa-sha2-512``, or ``default`` (which is the default). Ignored
        when not signing with an RSA key.

    not_before_duration
        Specifies the duration by which to backdate the ``ValidAfter`` property.
        Defaults to ``30s``.

    mount
        The name of the mount point the SSH secret backend is mounted at.
        Defaults to ``ssh``.
    """
    ret = {
        "name": name,
        "result": True,
        "comment": "The role is present as specified",
        "changes": {},
    }

    if not (allow_user_certificates or allow_host_certificates):
        ret["result"] = False
        ret["comment"] = "Either allow_user_certificates or allow_host_certificates must be true"
        return ret

    payload = {
        "default_user": default_user,
        "default_user_template": default_user_template,
        "allowed_users": allowed_users,
        "allowed_users_template": allowed_users_template,
        "allowed_domains": allowed_domains,
        "allowed_domains_template": allowed_domains_template,
        "ttl": ttl,
        "max_ttl": max_ttl,
        "allowed_critical_options": allowed_critical_options,
        "allowed_extensions": allowed_extensions,
        "default_critical_options": default_critical_options,
        "default_extensions": default_extensions,
        "default_extensions_template": default_extensions_template,
        "allow_user_certificates": allow_user_certificates,
        "allow_host_certificates": allow_host_certificates,
        "allow_bare_domains": allow_bare_domains,
        "allow_subdomains": allow_subdomains,
        "allow_user_key_ids": allow_user_key_ids,
        "key_id_format": key_id_format,
        "allowed_user_key_lengths": allowed_user_key_lengths,
        "algorithm_signer": algorithm_signer,
        "not_before_duration": not_before_duration,
    }
    return _role_present(name, "ca", ret, mount=mount, **payload)


def role_absent(name, mount="ssh"):
    """
    Ensure an SSH role is absent.

    name
        The name of the role.

    mount
        The name of the mount point the SSH secret backend is mounted at.
        Defaults to ``ssh``.
    """
    ret = {
        "name": name,
        "result": True,
        "comment": "The role is already absent",
        "changes": {},
    }

    try:
        try:
            __salt__["vault_ssh.read_role"](name, mount=mount)
        except CommandExecutionError as err:
            if "VaultNotFoundError" not in str(err):
                raise
            return ret

        ret["changes"]["deleted"] = name

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Role `{name}` would have been deleted."
            return ret

        __salt__["vault_ssh.delete_role"](name, mount=mount)

        try:
            __salt__["vault_ssh.read_role"](name, mount=mount)
        except CommandExecutionError as err:
            if "VaultNotFoundError" not in str(err):
                raise
            ret["comment"] = f"Role `{name}` has been deleted."
            return ret
        raise CommandExecutionError(
            "There were no errors during role deletion, but it is still reported as present."
        )

    except CommandExecutionError as err:
        ret["result"] = False
        ret["comment"] = str(err)
        ret["changes"] = {}

    return ret



"""
Class wrappers for several Vault (or OpenBao) API endpoints
"""

import salt.utils.json

from saltext.vault.utils.vault import leases
from saltext.vault.utils.vault.exceptions import VaultInvocationError
from saltext.vault.utils.vault.exceptions import VaultNotFoundError


class AppRoleApi:
    """
    Wraps the `Vault AppRole API <https://developer.hashicorp.com/vault/api-docs/auth/approle>`_.

    .. note::

        All durations can be specified either as an integer time in seconds
        or a time string like ``1h``.
    """

    def __init__(self, client):
        self.client = client

    def list_approles(self, mount="approle"):
        """
        List all AppRoles present on the specified mount.

        mount
            Name of the AppRole auth backend mount.
            Defaults to ``approle``.
        """
        endpoint = f"auth/{mount}/role"
        try:
            return self.client.list(endpoint)["data"]["keys"]
        except VaultNotFoundError:
            # When no AppRole has been created on the mount yet,
            # Vault returns 404 instead of an empty list.
            return []

    def read_approle(self, name, mount="approle"):
        """
        Read the properties of an existing AppRole.
        Raises VaultNotFound if the AppRole does not exist on the mount.

        name
            Name of the AppRole to read the properties of.

        mount
            Name of the AppRole auth backend mount.
            Defaults to ``approle``.
        """
        endpoint = f"auth/{mount}/role/{name}"
        return self.client.get(endpoint)["data"]

    def write_approle(
        self,
        name,
        bind_secret_id=None,
        secret_id_bound_cidrs=None,
        secret_id_num_uses=None,
        secret_id_ttl=None,
        local_secret_ids=None,
        token_ttl=None,
        token_max_ttl=None,
        token_policies=None,
        token_bound_cidrs=None,
        token_explicit_max_ttl=None,
        token_no_default_policy=None,
        token_num_uses=None,
        token_period=None,
        token_type=None,
        mount="approle",
    ):
        """
        Create or update an AppRole.

        name
            Name of the AppRole to read the properties of.

        bind_secret_id
            Require a SecretID when authenticating with this AppRole.
            Defaults to true.

        secret_id_bound_cidrs
            List of blocks of IP addresses in CIDR notation that
            can perform the login operation.

        secret_id_num_uses
            Number of times a generated SecretID can be used to authenticate
            with this AppRole by default. ``0`` means unlimited.

        secret_id_ttl
            Duration after which a generated SecretID for this AppRole expires by default.

        local_secret_ids
            If set, the secret IDs generated using this role will be cluster-local.
            This can only be set during role creation and once set, it can't be reset later.
            Defaults to false.

        token_ttl
            The incremental lifetime for tokens generated by authenticating with this AppRole.
            This value will be referenced at renewal time.

        token_max_ttl
            The maximum lifetime for tokens generated by authenticating with this AppRole.
            This value will be referenced at renewal time.

        token_policies
            List of token policies to encode onto generated tokens.
            This list may be supplemented by user/group/other values.

        token_bound_cidrs
            List of blocks of IP addresses in CIDR notation that
            can perform the login operation. The resulting token will be tied
            to these blocks as well.

        token_explicit_max_ttl
            Place a hard cap on the maximum lifetime of tokens issued by authenticating
            with this AppRole.

        token_no_default_policy
            Do not add the ``default`` policy to tokens generated by authenticating
            with this AppRole. Defaults to false.

        token_num_uses
            Number of times a token generated by authenticating with this AppRole
            may be used to issue requests. ``0`` means unlimited.

        token_period
            The maximum allowed period value when a periodic token is requested from this role.

        token_type
            The type of token that should be generated (``service``, ``batch`` or ``default``).

        mount
            Name of the AppRole auth backend mount.
            Defaults to ``approle``.
        """
        endpoint = f"auth/{mount}/role/{name}"
        payload = _filter_none(
            {
                "bind_secret_id": bind_secret_id,
                "secret_id_bound_cidrs": secret_id_bound_cidrs,
                "secret_id_num_uses": secret_id_num_uses,
                "secret_id_ttl": secret_id_ttl,
                "local_secret_ids": local_secret_ids,
                "token_ttl": token_ttl,
                "token_max_ttl": token_max_ttl,
                "token_policies": token_policies,
                "token_bound_cidrs": token_bound_cidrs,
                "token_explicit_max_ttl": token_explicit_max_ttl,
                "token_no_default_policy": token_no_default_policy,
                "token_num_uses": token_num_uses,
                "token_period": token_period,
                "token_type": token_type,
            }
        )
        return self.client.put(endpoint, payload=payload)

    def delete_approle(self, name, mount="approle"):
        """
        Delete an existing AppRole.
        Raises VaultNotFound if the AppRole does not exist on the mount.

        name
            Name of the AppRole to delete.

        mount
            Name of the AppRole auth backend mount.
            Defaults to ``approle``.
        """
        endpoint = f"auth/{mount}/role/{name}"
        return self.client.delete(endpoint)

    def read_role_id(self, name, wrap=False, mount="approle"):
        """
        Read the associated RoleID of an existing AppRole.
        Raises VaultNotFound if the AppRole does not exist on the mount.

        name
            Name of the AppRole.

        wrap
            If set, specifies the duration the resulting wrapping token should
            be valid for. This token can be used once to access the
            query result. Defaults to false (=> returns the RoleID as a string).

        mount
            Name of the AppRole auth backend mount.
            Defaults to ``approle``.
        """
        endpoint = f"auth/{mount}/role/{name}/role-id"
        role_id = self.client.get(endpoint, wrap=wrap)
        if wrap:
            return role_id
        return role_id["data"]["role_id"]

    def generate_secret_id(
        self,
        name,
        metadata=None,
        cidr_list=None,
        token_bound_cidrs=None,
        num_uses=None,
        ttl=None,
        wrap=False,
        mount="approle",
    ):
        """
        Generate a SecretID for an existing AppRole.
        Raises VaultNotFound if the AppRole does not exist on the mount.

        name
            Name of the AppRole.

        metadata
            Mapping of string keys to string values that specifies metadata
            to be set on the token generated by authenticating with this
            specific SecretID. It will be logged to audit logs in plaintext.

        cidr_list
            List of blocks of IP addresses in CIDR notation that
            can perform the login operation with this specific SecretID.
            If ``secret_id_bound_cidrs`` is set on the AppRole, this list
            must be a subset of the ones specified there.

        token_bound_cidrs
            List of blocks of IP addresses in CIDR notation that
            can perform the login operation. The resulting token will be tied
            to these blocks as well.
            If ``token_bound_cidrs`` is set on the AppRole, this list
            must be a subset of the ones specified there.

        num_uses
            Number of times this specific SecretID can be used to authenticate
            by default. ``0`` means unlimited.
            Must be equal to or lower than ``secret_id_num_uses`` set on the AppRole.

        ttl
            Duration after which this SecretID should expire.
            Must be equal to or lower than ``secret_id_ttl`` set on the AppRole.

        wrap
            If set, specifies the duration the resulting wrapping token should
            be valid for. This token can be used once to access the
            query result. Defaults to false (=> returns the SecretID as a string).

        mount
            Name of the AppRole auth backend mount.
            Defaults to ``approle``.
        """
        endpoint = f"auth/{mount}/role/{name}/secret-id"
        if metadata is not None:
            metadata = salt.utils.json.dumps(metadata)
        payload = _filter_none(
            {
                "metadata": metadata,
                "cidr_list": cidr_list,
                "token_bound_cidrs": token_bound_cidrs,
                "num_uses": num_uses,
                "ttl": ttl,
            }
        )
        response = self.client.post(endpoint, payload=payload, wrap=wrap)
        if wrap:
            return response
        # Sadly, secret_id_num_uses is not part of the information returned, but
        # it can be read with `read_secret_id` using the accessor.
        return leases.VaultSecretId(**response["data"])

    def read_secret_id(self, name, secret_id=None, accessor=None, mount="approle"):
        """
        Read properties of an existing SecretID.
        Raises VaultNotFound if the AppRole and/or SecretID does not exist on the mount.

        name
            Name of the AppRole the SecretID belongs to.

        secret_id
            The SecretID to look up. Specify either this or ``accessor``.

        accessor
            The accessor of the SecretID to look up. Specify either this
            or ``secret_id``.

        mount
            Name of the AppRole auth backend mount.
            Defaults to ``approle``.
        """
        if not secret_id and not accessor:
            raise VaultInvocationError("Need either secret_id or accessor to read secret ID.")
        if secret_id:
            endpoint = f"auth/{mount}/role/{name}/secret-id/lookup"
            payload = {"secret_id": str(secret_id)}
        else:
            endpoint = f"auth/{mount}/role/{name}/secret-id-accessor/lookup"
            payload = {"secret_id_accessor": accessor}
        try:
            return self.client.put(endpoint, payload=payload)["data"]
        except TypeError as err:
            # lookup does not raise exceptions, only returns True
            raise VaultNotFoundError() from err

    def destroy_secret_id(self, name, secret_id=None, accessor=None, mount="approle"):
        """
        Destroy an existing SecretID.
        Raises VaultNotFound if the AppRole and/or SecretID does not exist on the mount.

        name
            Name of the AppRole the SecretID belongs to.

        secret_id
            The SecretID to destroy. Specify either this or ``accessor``.

        accessor
            The accessor of the SecretID to destroy. Specify either this
            or ``secret_id``.

        mount
            Name of the AppRole auth backend mount.
            Defaults to ``approle``.
        """
        if not secret_id and not accessor:
            raise VaultInvocationError("Need either secret_id or accessor to destroy secret ID.")
        if secret_id:
            endpoint = f"auth/{mount}/role/{name}/secret-id/destroy"
            payload = {"secret_id": str(secret_id)}
        else:
            endpoint = f"auth/{mount}/role/{name}/secret-id-accessor/destroy"
            payload = {"secret_id_accessor": accessor}
        return self.client.post(endpoint, payload=payload)


class IdentityApi:
    """
    Wraps the Vault `Identity secret engine API <https://developer.hashicorp.com/vault/api-docs/secret/identity>`_.
    """

    def __init__(self, client):
        self.client = client

    def list_entities(self):
        """
        Return a list of the names of all entities known by Vault.
        """
        endpoint = "identity/entity/name"
        try:
            return self.client.list(endpoint)["data"]["keys"]
        except VaultNotFoundError:
            return []

    def read_entity(self, name):
        """
        Read the properties of an entity by its name.
        Raises VaultNotFound if the entity does not exist.

        name
            Name of the entity to read the properties of.
        """
        endpoint = f"identity/entity/name/{name}"
        return self.client.get(endpoint)["data"]

    def read_entity_by_alias(self, alias, mount):
        """
        Lookup the properties of an entity by its alias name and mount.
        Raises VaultNotFound if the entity does not exist.

        alias
            The name of the entity's alias on the specified
            ``mount``. For AppRole backends, this is the RoleID.

        mount
            The name of the mount the given alias is associated with.
            For example, if the backend is mounted at ``auth/approle``,
            this should be ``approle``.
        """
        endpoint = "identity/lookup/entity"
        payload = {
            "alias_name": alias,
            "alias_mount_accessor": self._lookup_mount_accessor(mount),
        }
        entity = self.client.put(endpoint, payload=payload)
        if isinstance(entity, dict):
            return entity["data"]
        raise VaultNotFoundError()

    def write_entity(self, name, metadata=None, policies=None, disabled=None):
        """
        Create or update an entity by name.

        name
            The name of the entity.

        metadata
            Mapping of string keys to string values that specifies metadata
            to be set on the entity. This can be used to template policies.

        policies
            List of policies to be tied to the entity. These policies will
            be active in addition to auth method-specific policies.

        disabled
            Whether this entity should be disabled. Disabled entities' associated
            tokens cannot be used, but are not revoked. Defaults to false.
        """
        endpoint = f"identity/entity/name/{name}"
        payload = _filter_none(
            {
                "metadata": metadata,
                "policies": policies,
                "disabled": disabled,
            }
        )
        return self.client.put(endpoint, payload=payload)

    def delete_entity(self, name):
        """
        Delete an entity by name.
        Raises VaultNotFound if the entity does not exist.

        name
            The name of the entity.
        """
        endpoint = f"identity/entity/name/{name}"
        return self.client.delete(endpoint)

    def write_entity_alias(self, name, alias_name, mount, custom_metadata=None):
        """
        Create/update the association between an entity and a specific
        alias of an auth mount.

        name
            Name of the entity to associate with the alias.

        alias_name
            Name of the alias to associate with the entity.
            The specifics are dependent on the type of the auth backend.
            For AppRoles, this is the RoleID.

        mount
            The name of the mount the given alias is associated with.
            For example, if the backend is mounted at ``auth/approle``,
            this should be ``approle``.

        custom_metadata
            A map of arbitrary string to string valued user-provided
            metadata meant to describe the alias.
        """
        entity = self.read_entity(name)
        mount_accessor = self._lookup_mount_accessor(mount)
        payload = {
            "canonical_id": entity["id"],
            "mount_accessor": mount_accessor,
            "name": alias_name,
        }
        if custom_metadata is not None:
            payload["custom_metadata"] = custom_metadata

        for alias in entity["aliases"]:
            # Ensure an existing alias is updated
            if alias["mount_accessor"] == mount_accessor:
                payload["id"] = alias["id"]
                break
        return self.client.put("identity/entity-alias", payload=payload)

    def _lookup_mount_accessor(self, mount):
        endpoint = f"sys/auth/{mount}"
        return self.client.get(endpoint)["data"]["accessor"]


def _filter_none(data):
    return {k: v for k, v in data.items() if v is not None}



"""
SSH wrapper for the :py:mod:`vault_pki <saltext.vault.modules.vault_pki>` execution module.

See there for documentation.

Setup notes
-----------
In addition to the regular :ref:`Vault setup <vault-setup>`, using
:py:func:`sign_certificate <saltext.vault.modules.vault_pki.sign_certificate>` and
:py:func:`revoke_certificate <saltext.vault.modules.vault_pki.revoke_certificate>`
requires the :py:mod:`x509_v2 <salt.modules.x509_v2>` module to be active on the target.

This means:

1. The global Python installation on the target needs to have the
   ``cryptography`` library installed.

2. On Salt releases below 3008, you need to include the following in your
   master configuration:

   .. code-block:: yaml

        # e.g. /etc/salt/master.d/salt_ssh.conf

        ssh_minion_opts:
          features:
            x509_v2: true
"""

from salt.utils.functools import namespaced_function

from saltext.vault.modules.vault_pki import _split_csr_kwargs
from saltext.vault.modules.vault_pki import _split_sans
from saltext.vault.modules.vault_pki import delete_issuer
from saltext.vault.modules.vault_pki import delete_key
from saltext.vault.modules.vault_pki import delete_role
from saltext.vault.modules.vault_pki import generate_root
from saltext.vault.modules.vault_pki import get_default_issuer
from saltext.vault.modules.vault_pki import issue_certificate
from saltext.vault.modules.vault_pki import list_certificates
from saltext.vault.modules.vault_pki import list_issuers
from saltext.vault.modules.vault_pki import list_revoked_certificates
from saltext.vault.modules.vault_pki import list_roles
from saltext.vault.modules.vault_pki import read_certificate
from saltext.vault.modules.vault_pki import read_issuer
from saltext.vault.modules.vault_pki import read_issuer_certificate
from saltext.vault.modules.vault_pki import read_issuer_crl
from saltext.vault.modules.vault_pki import read_role
from saltext.vault.modules.vault_pki import read_urls
from saltext.vault.modules.vault_pki import revoke_certificate
from saltext.vault.modules.vault_pki import set_default_issuer
from saltext.vault.modules.vault_pki import sign_certificate
from saltext.vault.modules.vault_pki import update_issuer
from saltext.vault.modules.vault_pki import write_role

_split_csr_kwargs = namespaced_function(_split_csr_kwargs, globals())
_split_sans = namespaced_function(_split_sans, globals())
delete_issuer = namespaced_function(delete_issuer, globals())
delete_key = namespaced_function(delete_key, globals())
delete_role = namespaced_function(delete_role, globals())
generate_root = namespaced_function(generate_root, globals())
get_default_issuer = namespaced_function(get_default_issuer, globals())
issue_certificate = namespaced_function(issue_certificate, globals())
list_certificates = namespaced_function(list_certificates, globals())
list_issuers = namespaced_function(list_issuers, globals())
list_revoked_certificates = namespaced_function(list_revoked_certificates, globals())
list_roles = namespaced_function(list_roles, globals())
read_certificate = namespaced_function(read_certificate, globals())
read_issuer = namespaced_function(read_issuer, globals())
read_issuer_certificate = namespaced_function(read_issuer_certificate, globals())
read_issuer_crl = namespaced_function(read_issuer_crl, globals())
read_role = namespaced_function(read_role, globals())
read_urls = namespaced_function(read_urls, globals())
revoke_certificate = namespaced_function(revoke_certificate, globals())
set_default_issuer = namespaced_function(set_default_issuer, globals())
sign_certificate = namespaced_function(sign_certificate, globals())
update_issuer = namespaced_function(update_issuer, globals())
write_role = namespaced_function(write_role, globals())



"""
SSH wrapper for the :py:mod:`vault_ssh <saltext.vault.modules.vault_ssh>` execution module.

See there for documentation.
"""

from salt.utils.functools import namespaced_function

from saltext.vault.modules.vault_ssh import _get_file_or_data
from saltext.vault.modules.vault_ssh import _write_role
from saltext.vault.modules.vault_ssh import create_ca
from saltext.vault.modules.vault_ssh import delete_role
from saltext.vault.modules.vault_ssh import delete_zeroaddr_roles
from saltext.vault.modules.vault_ssh import destroy_ca
from saltext.vault.modules.vault_ssh import generate_key_cert
from saltext.vault.modules.vault_ssh import get_creds
from saltext.vault.modules.vault_ssh import list_roles
from saltext.vault.modules.vault_ssh import list_roles_ip
from saltext.vault.modules.vault_ssh import list_roles_zeroaddr
from saltext.vault.modules.vault_ssh import read_ca
from saltext.vault.modules.vault_ssh import read_role
from saltext.vault.modules.vault_ssh import sign_key
from saltext.vault.modules.vault_ssh import write_role_ca
from saltext.vault.modules.vault_ssh import write_role_otp
from saltext.vault.modules.vault_ssh import write_zeroaddr_roles

_get_file_or_data = namespaced_function(_get_file_or_data, globals())
_write_role = namespaced_function(_write_role, globals())
create_ca = namespaced_function(create_ca, globals())
delete_role = namespaced_function(delete_role, globals())
delete_zeroaddr_roles = namespaced_function(delete_zeroaddr_roles, globals())
destroy_ca = namespaced_function(destroy_ca, globals())
generate_key_cert = namespaced_function(generate_key_cert, globals())
get_creds = namespaced_function(get_creds, globals())
list_roles = namespaced_function(list_roles, globals())
list_roles_ip = namespaced_function(list_roles_ip, globals())
list_roles_zeroaddr = namespaced_function(list_roles_zeroaddr, globals())
read_ca = namespaced_function(read_ca, globals())
read_role = namespaced_function(read_role, globals())
sign_key = namespaced_function(sign_key, globals())
write_role_ca = namespaced_function(write_role_ca, globals())
write_role_otp = namespaced_function(write_role_otp, globals())
write_zeroaddr_roles = namespaced_function(write_zeroaddr_roles, globals())



import pytest
from cryptography.hazmat.primitives import serialization
from salt.utils.x509 import NAME_ATTRS_OID
from salt.utils.x509 import generate_rsa_privkey
from salt.utils.x509 import load_cert

from tests.support.vault import vault_delete
from tests.support.vault import vault_disable_secret_engine
from tests.support.vault import vault_enable_secret_engine
from tests.support.vault import vault_list
from tests.support.vault import vault_read
from tests.support.vault import vault_write

pytest.importorskip("docker")

pytestmark = [
    pytest.mark.skip_if_binaries_missing("vault"),
    pytest.mark.usefixtures("container"),
]


@pytest.fixture(scope="module")
def minion_config_overrides(vault_port):
    return {
        "vault": {
            "auth": {
                "method": "token",
                "token": "testsecret",
            },
            "cache": {
                "backend": "disk",
            },
            "server": {
                "url": f"http://127.0.0.1:{vault_port}",
            },
        },
        "features": {"x509_v2": True},
    }


@pytest.fixture
def vault_pki(states):
    yield states.vault_pki


@pytest.fixture
def testrole():
    return {"ttl": 3600, "max_ttl": 86400, "allow_any_name": True, "enforce_hostnames": False}


@pytest.fixture
def ca2_cert():
    return """\
-----BEGIN CERTIFICATE-----
MIIDozCCAougAwIBAgIUGPU16um4LNbOXqUIEI5UjNOmgiUwDQYJKoZIhvcNAQEL
BQAwWDELMAkGA1UEBhMCVVMxCzAJBgNVBAgMAldBMRAwDgYDVQQHDAdTZWF0dGxl
MRIwEAYDVQQKDAlTYWx0U3RhY2sxFjAUBgNVBAMMDVRlc3QgUmVpc3N1ZXIwIBcN
MjQwNzIzMDgwMzM5WhgPMjA1NDA3MjQwODAzMzlaMFgxCzAJBgNVBAYTAlVTMQsw
CQYDVQQIDAJXQTEQMA4GA1UEBwwHU2VhdHRsZTESMBAGA1UECgwJU2FsdFN0YWNr
MRYwFAYDVQQDDA1UZXN0IFJlaXNzdWVyMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8A
MIIBCgKCAQEAueJiKvUtqz429f+7AZ9X092L/siLlSukAUxCI+E/Zz37pXvRcQi9
50rsgxdUKG/5epJX46oxxWlW2CyWwXOCWvWr7CNe8wMOrNqi2Et33PNAnyUe9+iX
tfdQ33RCdrrAUVI7IUiM+WXkSqgaFCke7IdFA0FXa6+v1bkgfhwETsxelLrWpM9d
oBOh5mZLIjYjbAlTnKHemNqXYlJvgqtFq6s+KZ4tlX9f1WZkOghORPkAvti7VBFO
0uz0UMETBszlYlPVODw3DYdJrOlq4cjl7wxNnzNilAaRx2p7PiHDlFAROMAgrufq
7RDw/l5pL6vJbPC6+wu/UzWthPZx9mBGRQIDAQABo2MwYTAdBgNVHQ4EFgQUALDS
25ITRPYLJ6itcwFQ1gKprtYwHwYDVR0jBBgwFoAUALDS25ITRPYLJ6itcwFQ1gKp
rtYwDwYDVR0TAQH/BAUwAwEB/zAOBgNVHQ8BAf8EBAMCAYYwDQYJKoZIhvcNAQEL
BQADggEBAH7TMlojvSQOO04RyayzGddiugElad30226G2RYEE6hUGA/wuRmf3UlV
0FddU+7vaEwaTXJKtjchI/MZ6yFZpNhXRWDnSo1jGIXZSxSXYkAjRI0tIE3Vt/Qs
ySmkDvfb/BtXCCinBr1833DuKF8GAbnLhoR6yHx6HFhYjMjiwgIuldw21D4skpjQ
h9bkSYnj8lsoz8m2JEbXYag+vHaVGHJ6mPFPKQWG1CWko+ONwSdXZO7nVOpk2JJm
vfAVwCW9ly5eg8M+nIBjxoDGxgiVweuxe7kfMhOKvBZJ9UmGTnOkHZR328cBliMd
wUKPgGL2SQ47Iyzegf2FmSv+wgvGUpI=
-----END CERTIFICATE-----
"""


@pytest.fixture
def ca2_key():
    return """\
-----BEGIN PRIVATE KEY-----
MIIEvQIBADANBgkqhkiG9w0BAQEFAASCBKcwggSjAgEAAoIBAQC54mIq9S2rPjb1
/7sBn1fT3Yv+yIuVK6QBTEIj4T9nPfule9FxCL3nSuyDF1Qob/l6klfjqjHFaVbY
LJbBc4Ja9avsI17zAw6s2qLYS3fc80CfJR736Je191DfdEJ2usBRUjshSIz5ZeRK
qBoUKR7sh0UDQVdrr6/VuSB+HAROzF6Uutakz12gE6HmZksiNiNsCVOcod6Y2pdi
Um+Cq0Wrqz4pni2Vf1/VZmQ6CE5E+QC+2LtUEU7S7PRQwRMGzOViU9U4PDcNh0ms
6WrhyOXvDE2fM2KUBpHHans+IcOUUBE4wCCu5+rtEPD+Xmkvq8ls8Lr7C79TNa2E
9nH2YEZFAgMBAAECggEAL720+NN/pzuTYhsMLJ6AMCn2irl3IBjVRoAPfKedYSbK
OvZSFHXqUD0uAX08YCZiLNjpOc+8eLdVVrAdCBJiqHnwbfWnrUJbwolkyaiYYGcZ
ccZW7dUPIe0jGEED0Pql6jz0ctfvXR6OQ55pFER7bMRRNUTS7xVwU7P4ZGtNr+7c
21Tu5X6THIUDy2PjRianENQJZ2GEsPQe4Sh0sieZvrIf1+yN9rpJJRjRj2oTWyBk
RXw37us828XqG3obe9jkDDlSr4+IWCKAIebzRcd++WnuIRIAdE51HBrYEiZYWU+Y
LjmpchAMZdiOLfa1ARehnn8ElyxXPH7iR2AjdO3dAQKBgQDuisAcK7oOeUHJiHoB
c9vYJmavPBjCC1A0gIlFixRu+78GIhkTCwsoLA6hKH4nqNzeztcJixaHdRmuKXPt
HJz13yJw/nQ1pqd0MBZnPwMMaibTbAUbjLwxVGi00zfdzNQfCDDkIMIHQ9GxuXcz
w/977Jb6dC4EEx0e6ZAfZk+LhwKBgQDHfQ4KYA6bowvBusOrWn2fHNPivk0Ql2I8
mKuhSeWAcmtEh9Fsver44zqz2xi1nj41zXGsQxdQgcxe9CJ06WQDY5kS0CKnP0mQ
T3RGHnjhStg40N3zOLTRsBZlXkikkVctmWnjT/NL94d1rfRN4UKmOr5zSr7Zxw9U
G9mA9vsK0wKBgQDg7dqqdZzyaupaw6Lv3bTOg59N22gpCQvvBcjq13NEF4QPn3Vv
XHl/vtNoqUsT0Im8WuOv7wQmZIf7jsDuM43Z1jaev4EK2gOKbpGhd2xDd5D2ySOj
z7fg+AvnfkdukObwAARCCJWMzilb4VuCZ21wCC5xKb3+P5u0+13YDdwx3wKBgF8r
uCEXBqEVviwkj+kV+MyKEkQgid+aeVFzfJ4sBQOskqRVL4JzcMBgl8bqhfVPk1pT
syF9uIe+BORgEHg6SG6de4/QIFguB0iDv3McYosJC/K/IsRAj3NiUKz3uCxa8n5c
rHm30NizNLrdzKnDB+sKJ4YVaMu4/gUgbDnsmoPlAoGAIcb4wu38jH6Ynz31s2l4
bGvNummvXmwlX8EllmuyOh6/0W209o2vAQ+fz3vvhtnyBE9rfZLaQDWN4eb4cDEV
xeM1Z86IbwIf8HQVmxwZMzR/qFPACXxR9uq79Gp4817ZFXRBPBmqp6X1Pa7VbNg0
tClJiP0NZQ8YBJ+vi2VB1iQ=
-----END PRIVATE KEY-----
"""


@pytest.fixture
def ca_cert():
    return """\
-----BEGIN CERTIFICATE-----
MIIDODCCAiCgAwIBAgIIbfpgqP0VGPgwDQYJKoZIhvcNAQELBQAwKzELMAkGA1UE
BhMCVVMxDTALBgNVBAMMBFRlc3QxDTALBgNVBAoMBFNhbHQwHhcNMjIxMTE1MTQw
NDMzWhcNMzIxMTEyMTQwNDMzWjArMQswCQYDVQQGEwJVUzENMAsGA1UEAwwEVGVz
dDENMAsGA1UECgwEU2FsdDCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEB
AOGTScvrjcEt6vsJcG9RUp6fKaDNDWZnJET0omanK9ZwaoGpJPp8UDYe/8ADeI7N
10wdyB4oDM9gRDjInBtdQO/PsrmKZF6LzqVFgLMxu2up+PHMi9z6B2P4esIAzMu9
PYxc9zH4HzLImHqscVD2HCabsjp9X134Af7hVY5NN/W/4qTP7uOM20wSG2TPI6+B
tA9VyPbEPMPRzXzrqc45rVYe6kb2bT84GE93Vcu/e5JZ/k2AKD8Hoa2cxLPsTLq5
igl+D+k+dfUtiABiKPvVQiYBsD1fyHDn2m7B6pCgvrGqHjsoAKufgFnXy6PJRg7n
vQfaxSiusM5s+VS+fjlvgwsCAwEAAaNgMF4wDwYDVR0TBAgwBgEB/wIBATALBgNV
HQ8EBAMCAQYwHQYDVR0OBBYEFFzy8fRTKSOe7kBakqO0Ki71potnMB8GA1UdIwQY
MBaAFFzy8fRTKSOe7kBakqO0Ki71potnMA0GCSqGSIb3DQEBCwUAA4IBAQBZS4MP
fXYPoGZ66seM+0eikScZHirbRe8vHxHkujnTBUjQITKm86WeQgeBCD2pobgBGZtt
5YFozM4cERqY7/1BdemUxFvPmMFFznt0TM5w+DfGWVK8un6SYwHnmBbnkWgX4Srm
GsL0HHWxVXkGnFGFk6Sbo3vnN7CpkpQTWFqeQQ5rHOw91pt7KnNZwc6I3ZjrCUHJ
+UmKKrga16a4Q+8FBpYdphQU609npo/0zuaE6FyiJYlW3tG+mlbbNgzY/+eUaxt2
9Bp9mtA+Hkox551Mfpq45Oi+ehwMt0xjZCjuFCM78oiUdHCGO+EmcT7ogiYALiOF
LN1w5sybsYwIw6QN
-----END CERTIFICATE-----
"""


@pytest.fixture
def ca_key():
    return """\
-----BEGIN RSA PRIVATE KEY-----
MIIEowIBAAKCAQEA4ZNJy+uNwS3q+wlwb1FSnp8poM0NZmckRPSiZqcr1nBqgakk
+nxQNh7/wAN4js3XTB3IHigMz2BEOMicG11A78+yuYpkXovOpUWAszG7a6n48cyL
3PoHY/h6wgDMy709jFz3MfgfMsiYeqxxUPYcJpuyOn1fXfgB/uFVjk039b/ipM/u
44zbTBIbZM8jr4G0D1XI9sQ8w9HNfOupzjmtVh7qRvZtPzgYT3dVy797kln+TYAo
PwehrZzEs+xMurmKCX4P6T519S2IAGIo+9VCJgGwPV/IcOfabsHqkKC+saoeOygA
q5+AWdfLo8lGDue9B9rFKK6wzmz5VL5+OW+DCwIDAQABAoIBAFfImc9hu6iR1gAb
jEXFwAE6r1iEc9KGEPdEvG52X/jzhn8u89UGy7BEIAL5VtE8Caz1agtSSqnpLKNs
blO31q18hnDuCmFAxwpKIeuaTvV3EAoJL+Su6HFfIWaeKRSgcHNPOmOXy4xXw/75
XJ/FJu9fZ9ybLaHEAgLObh0Sr9RSPQbZ72ZawPP8+5WCbR+2w90RApHXQL0piSbW
lIx1NE6o5wQb3vik8z/k5FqLCY2a8++WNyfvS+WWFY5WXGI7ZiDDQk46gnslquH2
Lon5CEn3JlTGQFhxaaa2ivssscf2lA2Rvm2E8o1rdZJS2OpSE0ai4TXY9XnyjZj1
5usWIwECgYEA+3Mwu03A7PyLEBksS/u3MSo/176S9lF/uXcecQNdhAIalUZ8AgV3
7HP2yI9ZC0ekA809ZzFjGFostXm9VfUOEZ549jLOMzvBtCdaI0aBUE8icu52fX4r
fT2NY6hYgz5/fxD8sq1XH/fqNNexABwtViH6YAly/9A1/8M3BOWt72UCgYEA5ag8
sIfiBUoWd1sS6qHDuugWlpx4ZWYC/59XEJyCN2wioP8qFji/aNZxF1wLfyQe/zaa
YBFusjsBnSfBU1p4UKCRHWQ9/CnC0DzqTkyKC4Fv8GuxgywNm5W9gPKk7idHP7mw
e+7Uvf1pOQccqEPh7yltpW+Xw27gfsC2DMAIGa8CgYByv/q5P56PiCCeVB6W/mR3
l2RTPLEsn7y+EtJdmL+QgrVG8kedVImJ6tHwbRqhvyvmYD9pXGxwrJZCqy/wjkjB
WaSyFjVrxBV99Yd5Ga/hyntaH+ELHA0UtoZTuHvMSTU9866ei+R6vlSvkM9B0ZoO
+KqeMTG99HLwKVJudbKO0QKBgQCd33U49XBOqoufKSBr4yAmUH2Ws6GgMuxExUiY
xr5NUyzK+B36gLA0ZZYAtOnCURZt4x9kgxdRtnZ5jma74ilrY7XeOpbRzfN6KyX3
BW6wUh6da6rvvUztc5Z+Gk9+18mG6SOFTr04jgfTiCwPD/s06YnSfFAbrRDukZOU
WD45SQKBgBvjSwl3AbPoJnRjZjGuCUMKQKrLm30xCeorxasu+di/4YV5Yd8VUjaO
mYyqXW6bQndKLuXT+AXtCd/Xt2sI96z8mc0G5fImDUxQjMUuS3RyQK357cEOu8Zy
HdI7Pfaf/l0HozAw/Al+LXbpmSBdfmz0U/EGAKRqXMW5+vQ7XHXD
-----END RSA PRIVATE KEY-----"""


@pytest.fixture
def ca_sub_key():
    return """\
-----BEGIN PRIVATE KEY-----
MIIEvQIBADANBgkqhkiG9w0BAQEFAASCBKcwggSjAgEAAoIBAQCqbHMHARaW7ncu
bcgLhHEyo9wEOh4a82D5B1Y1VH/XJvwBD4Bt59HS46zqYNtXXRbWnWkISL8wvWmC
JEXZ7GmoTL+e8s/h8C1A0vqeThBv0TJyZK93CRD80vG/v+NiLd2SjlPknXibDwFz
wrkZnGLryRqmswvG12ahOYQYWlPCv+BIUxDUL/Pz0z4ZEgoZgArVXMiqXlCfvsPy
Uenx5VLNQZB95K88TMhr/fCbrncddlvDzUPlW4IpLbyfKE9MCCiWnXYqgMr/WzJx
6u6Pf9B4G+2GVCYNEeNFcsbYOgRuZAz5tSrIvIaiWqz0Uy8/+7xRHV/80EO4eUGA
aNJUWRoxAgMBAAECggEAIVNetPZuA+qyzJX0IehuuE/ZlMwGmg+QnXHlVj1lWF3L
tqtg2l0UJ1CVPindinpuHl6erNuI44+Og7/zFtfHm30SlZL2usBcIQqArpcmWK9I
VZ1BwJ25wC7BzlTIMqk0ZFXHqvNuI6guCQSBbLQrld74ArQNb/8sFwfnwFlder33
cEEkxzkz1+tjDVPllJqhDHC4pWHfGiU/NzfLnMKTnlViOYHywYLPKi4T5H4PA/Nd
Lwqnv4AURmBXxi0pYc/7640pZiBuyNfAW6zik4m4hZMPdznpRyLC2VAZMZWLdxry
38a1HHK4ZgQtq7hgS10GtJLqoB/hfFQPVTLdKKg2DwKBgQDqmrFaL7UodukHoTQT
Ob7WmizCCRRjYfljREJMHa7DgLODjVfrBcHWgKV6sB8OFESx+FJA9IW8yOaGwmKJ
m2PlNTeAXt6wfLdX2pq+8Wb4PFmYeqxrAubItYdiQEnVYZOWCpTx+aByHz76F5if
DHXvaeYztyjE7b88e+gIWv1fjwKBgQC591SvojZ57J1yQGUutNese/QqqhFbBJz9
RWCm217ksoqVZYZWxsG+wGH6R0pDjRYxn0wXrrqCLag4TI73qPawUV22C/+m7L2o
mmJ4BXxklmrtYBXd/xdMTOvzvxskClbK1oYmloWGh4LmZFBKq7DTsWXzyidZJz5i
1rwM0A0KPwKBgQCKEs8sgAWDsjBF8EdAxWyeyxBqhoN8Vk47cRH/0DxqDZYZZ5eF
19aUUxSRV5R/achgYgCu//qx+B9M0pzB1jV90cs/fxZbEpupVhxbIqJymLo2doSB
WqzPFZ9/YMzTi+EbnlC49SzL3b3n3PlTKjdC17XHXBXfiPlTNK2ENWEH2wKBgFHE
fmf7WxihAVmLFvJCcdJVbjaUMK1kieKS7rxvGHpWRrkJutfM7MOCs5HoZq7tCiUn
db20Bi3XBXA7uWEL2ewM2reA7xfmYD4SI9nCD7/qo3lcFkFWOFhEOjsifDyMjz0A
tluhM3TDgLrswKEUfNuX1MwsxsBckQHEiUrY7+LhAoGABk3BBSVayHrTCT4xJipk
PIlGdKHjzdnI4adfJEWlGQA5EUv+fiorfeFW0u1yKyiftbCrkwG8z4S5JB3PkBZc
kbVyt+nUh7+gOoNE0ebGJzirkLiFzeulFmQFE3VzwxXuuBm/Zl8ruOiRjftooJx0
A1vWs/C545K2v/LLxc55Cug=
-----END PRIVATE KEY-----
"""


@pytest.fixture
def ca_sub_cert():
    return """\
-----BEGIN CERTIFICATE-----
MIIDTTCCAjWgAwIBAgIUJ8/bURqv1pOka8sHUVqp+C4+FWIwDQYJKoZIhvcNAQEL
BQAwKzELMAkGA1UEBhMCVVMxDTALBgNVBAMMBFRlc3QxDTALBgNVBAoMBFNhbHQw
HhcNMjUwNTA0MTQyMjUxWhcNMzUwNTAyMTQyMjUxWjAuMQswCQYDVQQGEwJVUzEN
MAsGA1UECgwEU2FsdDEQMA4GA1UEAwwHVGVzdFN1YjCCASIwDQYJKoZIhvcNAQEB
BQADggEPADCCAQoCggEBAKpscwcBFpbudy5tyAuEcTKj3AQ6HhrzYPkHVjVUf9cm
/AEPgG3n0dLjrOpg21ddFtadaQhIvzC9aYIkRdnsaahMv57yz+HwLUDS+p5OEG/R
MnJkr3cJEPzS8b+/42It3ZKOU+SdeJsPAXPCuRmcYuvJGqazC8bXZqE5hBhaU8K/
4EhTENQv8/PTPhkSChmACtVcyKpeUJ++w/JR6fHlUs1BkH3krzxMyGv98Juudx12
W8PNQ+VbgiktvJ8oT0wIKJaddiqAyv9bMnHq7o9/0Hgb7YZUJg0R40Vyxtg6BG5k
DPm1Ksi8hqJarPRTLz/7vFEdX/zQQ7h5QYBo0lRZGjECAwEAAaNmMGQwEgYDVR0T
AQH/BAgwBgEB/wIBADAOBgNVHQ8BAf8EBAMCAQYwHQYDVR0OBBYEFDru5IneLp7q
OrwVwHw6i3deyooOMB8GA1UdIwQYMBaAFFzy8fRTKSOe7kBakqO0Ki71potnMA0G
CSqGSIb3DQEBCwUAA4IBAQBJQJcAMsbKJObyn9uX5JTy/pFptden0c9XXmwdNq53
fY85pWWSN4E5880yWzhVtB60z4hR4V0hI07928rx+zqgYSRvFJD4Sv50ju7QzjtK
cMm0oySqBACJBoQvQjffpnFxsMiPVpbVuEDmDYGrlAkAaXy9O/AMD5D3476QuMsV
4IPOZ65ISOE7yTlYsIXMcEvAej8Rv1uRSScWwoxU7F0XsULMXMfVdW6b0/x/Js2N
UfxmAJVK8gpR1J2uT0LZgZ5QHgGagYtDwiWYyW/w5fSzxCA43KrOg6g4x+y/PBFj
cYpgWHc7NNeYGs6uKgA+IJJalICKGMSJpStncc6SGeKi
-----END CERTIFICATE-----
"""


@pytest.fixture
def cert_args(tmp_path, private_key):
    return {
        "name": f"{tmp_path}/cert",
        "common_name": "saltproject.io",
        "role_name": "testrole",
        "private_key": private_key,
        "ttl": "30m",
        "ttl_remaining": 0,
    }


@pytest.fixture(scope="module", autouse=True)
def pki_engine(container):  # pylint: disable=unused-argument
    assert vault_enable_secret_engine("pki")
    yield
    assert vault_disable_secret_engine("pki")


@pytest.fixture(scope="module")
def private_key():
    pk = generate_rsa_privkey(2048)
    pk_bytes = pk.private_bytes(
        serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )
    return pk_bytes.decode()


@pytest.fixture(params=[["testrole"]])
def roles_setup(request):  # pylint: disable=unused-argument
    try:
        for role_name in request.param:
            role_args = request.getfixturevalue(role_name)
            vault_write(f"pki/roles/{role_name}", **role_args)
            assert role_name in vault_list("pki/roles")
        yield
    finally:
        for role_name in request.param:
            if role_name in vault_list("pki/roles"):
                vault_delete(f"pki/roles/{role_name}")
                assert role_name not in vault_list("pki/roles")


@pytest.fixture
def issuer_setup(ca_cert, ca_key):
    ret_data = vault_write("/pki/config/ca", pem_bundle="\n".join([ca_cert, ca_key]))["data"]
    issuer_id = ret_data["imported_issuers"][0]
    vault_write(f"/pki/issuer/{issuer_id}", issuer_name="root")
    try:
        yield issuer_id
    finally:
        vault_delete(f"/pki/issuer/{issuer_id}")


@pytest.fixture
def issuer_setup_additional(ca2_cert, ca2_key):
    ret_data = vault_write("/pki/config/ca", pem_bundle="\n".join([ca2_cert, ca2_key]))["data"]
    issuer_id = ret_data["imported_issuers"][0]
    vault_write(f"/pki/issuer/{issuer_id}", issuer_name="additional")
    try:
        yield issuer_id
    finally:
        vault_delete(f"/pki/issuer/{issuer_id}")


@pytest.fixture
def issuer_setup_sub(ca_cert, ca_sub_cert, ca_sub_key):
    ret_data = vault_write(
        "/pki/config/ca", pem_bundle="\n".join([ca_sub_cert, ca_sub_key, ca_cert])
    )["data"]
    issuers = ret_data["mapping"]
    imported_key = ret_data["imported_keys"][0]
    try:
        root_id = next(x for x in issuers if issuers[x] != imported_key)
        sub_id = next(x for x in issuers if issuers[x] if issuers[x] == imported_key)
    except StopIteration as err:
        raise AssertionError("Unable to find issuer IDs") from err
    vault_write(f"/pki/issuer/{sub_id}", issuer_name="sub")
    try:
        yield sub_id
    finally:
        vault_delete(f"/pki/issuer/{sub_id}")
        vault_delete(f"/pki/issuer/{root_id}")


@pytest.mark.usefixtures("issuer_setup")
@pytest.mark.usefixtures("roles_setup")
def test_certificate_managed_create(vault_pki, cert_args):
    ret = vault_pki.certificate_managed(**cert_args)
    assert ret.result
    assert ret.changes
    assert "created" in ret.changes


@pytest.mark.usefixtures("issuer_setup")
@pytest.mark.usefixtures("roles_setup")
def test_certificate_managed_state_no_changes(vault_pki, cert_args):
    ret = vault_pki.certificate_managed(**cert_args)
    assert ret.result
    assert ret.changes
    assert "created" in ret.changes

    # Try again
    ret = vault_pki.certificate_managed(**cert_args)
    assert ret.result
    assert not ret.changes


@pytest.mark.usefixtures("issuer_setup")
@pytest.mark.usefixtures("roles_setup")
def test_certificate_managed_is_reissued_forcibly(vault_pki, cert_args):
    ret = vault_pki.certificate_managed(**cert_args)
    assert "created" in ret.changes

    cert_args["reissue"] = True
    ret = vault_pki.certificate_managed(**cert_args)
    assert "replaced" in ret.changes


@pytest.mark.usefixtures("issuer_setup")
@pytest.mark.usefixtures("roles_setup")
@pytest.mark.parametrize("encoding", ["der", "pem"])
def test_certificate_managed_encoding(vault_pki, cert_args, encoding):
    cert_args["encoding"] = encoding
    ret = vault_pki.certificate_managed(**cert_args)
    assert ret.result

    _, enc, _, _ = load_cert(cert_args["name"], get_encoding=True)

    assert enc == encoding


@pytest.mark.usefixtures("issuer_setup_sub")
@pytest.mark.usefixtures("roles_setup")
def test_certificate_managed_includes_chain(vault_pki, cert_args):
    cert_args["append_ca_chain"] = True
    cert_args["issuer_ref"] = "sub"
    ret = vault_pki.certificate_managed(**cert_args)
    assert ret.result
    assert ret.changes

    _, chain = load_cert(cert_args["name"], load_chain=True)

    assert len(chain) == 1

    # Ensure it's idempotent still
    ret = vault_pki.certificate_managed(**cert_args)
    assert ret.result
    assert not ret.changes


@pytest.mark.usefixtures("issuer_setup")
@pytest.mark.usefixtures("roles_setup")
@pytest.mark.parametrize(
    "attr",
    [
        ({"L": "Boston"}),
        ({"C": "US"}),
        ({"ST": "That Street"}),
        ({"O": "Salt Project"}),
        ({"OU": "Salt Extensions"}),
    ],
)
def test_certificate_managed_sign_verbatim(vault_pki, cert_args, attr):
    cert_args = {**cert_args, **attr}
    cert_args["sign_verbatim"] = True
    ret = vault_pki.certificate_managed(**cert_args)
    assert ret.result

    cert = load_cert(cert_args["name"])

    for k, v in attr.items():
        c_attrs = cert.subject.get_attributes_for_oid(NAME_ATTRS_OID[k])
        assert len(c_attrs) == 1
        assert c_attrs[0].value == v


@pytest.mark.usefixtures("issuer_setup")
@pytest.mark.usefixtures("roles_setup")
def test_certificate_managed_changed_cn(vault_pki, cert_args):
    ret = vault_pki.certificate_managed(**cert_args)
    assert ret.result
    assert "created" in ret.changes

    cert_args["common_name"] = "brand new common name"
    ret = vault_pki.certificate_managed(**cert_args)
    cert = load_cert(cert_args["name"])

    assert "subject" in ret.changes
    assert "CN" in ret.changes["subject"]

    c_attrs = cert.subject.get_attributes_for_oid(NAME_ATTRS_OID["CN"])
    assert c_attrs[0].value == "brand new common name"


@pytest.mark.usefixtures("issuer_setup")
@pytest.mark.usefixtures("roles_setup")
@pytest.mark.parametrize(
    "attr,replace",
    [
        ({"L": "Boston"}, {"L": "Moscow"}),
        ({"C": "US"}, {"C": "RU"}),
        ({"ST": "That Street"}, {"ST": "Other Street"}),
        ({"O": "Salt Project"}, {"O": "Salt"}),
        ({"OU": "Salt Extensions"}, {"OU": "Extensions"}),
    ],
)
def test_certificate_managed_changed_subject(vault_pki, cert_args, attr, replace):
    cert_args["sign_verbatim"] = True
    cert_args = {**cert_args, **attr}
    ret = vault_pki.certificate_managed(**cert_args)
    assert ret.result
    assert "created" in ret.changes

    cert_args = {**cert_args, **replace}
    ret = vault_pki.certificate_managed(**cert_args)
    cert = load_cert(cert_args["name"])

    assert "subject" in ret.changes

    for k, v in replace.items():
        assert k in ret.changes["subject"]
        c_attrs = cert.subject.get_attributes_for_oid(NAME_ATTRS_OID[k])
        assert len(c_attrs) == 1
        assert c_attrs[0].value == v


@pytest.mark.usefixtures("issuer_setup")
@pytest.mark.usefixtures("issuer_setup_additional")
@pytest.mark.usefixtures("roles_setup")
def test_certificate_managed_changed_issuer(vault_pki, cert_args):
    cert_args["issuer_ref"] = "root"
    ret = vault_pki.certificate_managed(**cert_args)
    assert ret.result
    assert "created" in ret.changes

    cert_args["issuer_ref"] = "additional"
    ret = vault_pki.certificate_managed(**cert_args)
    assert ret.result
    assert "issuer_name" in ret.changes


@pytest.mark.usefixtures("issuer_setup")
def test_role_managed(vault_pki):
    ret = vault_pki.role_managed("dummy")
    assert ret.result
    assert "created" in ret.changes

    ret = vault_pki.role_managed("dummy")
    assert ret.result
    assert not ret.changes


@pytest.mark.usefixtures("issuer_setup")
@pytest.mark.usefixtures("issuer_setup_additional")
@pytest.mark.parametrize("issuer_ref", ["additional", "root"])
def test_role_managed_correct_issuer(vault_pki, issuer_ref):
    ret = vault_pki.role_managed("dummy", issuer_ref=issuer_ref)
    assert ret.result

    role_info = vault_read("pki/roles/dummy")["data"]
    assert role_info["issuer_ref"] == issuer_ref


@pytest.mark.usefixtures("issuer_setup")
@pytest.mark.usefixtures("roles_setup")
@pytest.mark.parametrize(
    "params",
    [
        {
            "allow_localhost": False,
            "allow_bare_domains": True,
            "allowed_domains": ["www.example.com", "www.acme.com"],
            "allow_subdomains": True,
            "allow_glob_domains": True,
        },
        {"server_flag": False, "client_flag": False, "no_store": True},
        {"organization": ["Salt"], "country": ["US"], "locality": ["Seattle"], "require_cn": False},
    ],
)
def test_role_managed_payload(vault_pki, params):
    ret = vault_pki.role_managed("testrole", **params)

    role_info = vault_read("pki/roles/testrole")["data"]

    for k, v in params.items():
        assert ret.changes[k]["new"] == v
        assert role_info[k] == v


@pytest.mark.usefixtures("issuer_setup")
@pytest.mark.usefixtures("roles_setup")
@pytest.mark.parametrize(
    "ttl,expected", [(60, 60), ("10m", 600), ("1h", 3600), ("1d", 86400), ("30d", 2592000)]
)
def test_role_managed_ttl(vault_pki, ttl, expected):
    vault_pki.role_managed("testrole", ttl=ttl, max_ttl="365d")

    role_info = vault_read("pki/roles/testrole")["data"]
    assert role_info["ttl"] == expected


@pytest.mark.usefixtures("issuer_setup")
@pytest.mark.usefixtures("roles_setup")
@pytest.mark.parametrize(
    "max_ttl,expected", [(60, 60), ("10m", 600), ("1h", 3600), ("1d", 86400), ("30d", 2592000)]
)
def test_role_managed_max_ttl(vault_pki, max_ttl, expected):
    vault_pki.role_managed("testrole", ttl=1, max_ttl=max_ttl)

    role_info = vault_read("pki/roles/testrole")["data"]
    assert role_info["max_ttl"] == expected


@pytest.mark.usefixtures("roles_setup")
def test_role_absent(vault_pki):
    assert "testrole" in vault_list("pki/roles")
    ret = vault_pki.role_absent("testrole")
    assert "deleted" in ret.changes
    assert "testrole" not in vault_list("pki/roles")
